import cellrank.tl.kernels
import cellrank.tl.estimators
from cellrank.tl._lineage import Lineage
from cellrank.tl._lineages import lineages, lineage_drivers
from cellrank.tl._init_term_states import initial_states, terminal_states
from cellrank.tl._transition_matrix import transition_matrix



from typing import Optional, Sequence

import pytest

import cellrank as cr
from anndata import AnnData
from cellrank.tl._constants import AbsProbKey, TermStatesKey, _probs
from cellrank.tl.kernels._base_kernel import KernelAdd

import numpy as np
import pandas as pd


class TestLineages:
    def test_no_root_cells(self, adata: AnnData):
        with pytest.raises(RuntimeError):
            cr.tl.lineages(adata)

    def test_normal_run(self, adata_cflare):
        cr.tl.lineages(adata_cflare)

        assert str(AbsProbKey.FORWARD) in adata_cflare.obsm.keys()

    def test_normal_run_copy(self, adata_cflare):
        adata_cr2 = cr.tl.lineages(adata_cflare, copy=True)

        assert isinstance(adata_cr2, AnnData)
        assert adata_cflare is not adata_cr2


class TestLineageDrivers:
    def test_no_abs_probs(self, adata: AnnData):
        with pytest.raises(RuntimeError):
            cr.tl.lineage_drivers(adata)

    def test_normal_run(self, adata_cflare):
        cr.tl.lineages(adata_cflare)
        cr.tl.lineage_drivers(adata_cflare, use_raw=False)

        for name in adata_cflare.obsm[AbsProbKey.FORWARD.s].names:
            assert np.all(adata_cflare.var[f"to {name} corr"] >= -1.0)
            assert np.all(adata_cflare.var[f"to {name} corr"] <= 1.0)

            assert np.all(adata_cflare.var[f"to {name} qval"] >= 0)
            assert np.all(adata_cflare.var[f"to {name} qval"] <= 1.0)

    def test_normal_run_raw(self, adata_cflare):
        adata_cflare.raw = adata_cflare.copy()
        cr.tl.lineages(adata_cflare)
        cr.tl.lineage_drivers(adata_cflare, use_raw=True)

        for name in adata_cflare.obsm[AbsProbKey.FORWARD.s].names:
            assert np.all(adata_cflare.raw.var[f"to {name} corr"] >= -1.0)
            assert np.all(adata_cflare.raw.var[f"to {name} corr"] <= 1.0)

            assert np.all(adata_cflare.raw.var[f"to {name} qval"] >= 0)
            assert np.all(adata_cflare.raw.var[f"to {name} qval"] <= 1.0)

    @pytest.mark.parametrize("lineages", [None, ["0"], ["0, 1"]])
    def test_return_drivers(self, adata_cflare, lineages: Optional[Sequence[str]]):
        cr.tl.terminal_states(adata_cflare, n_states=3)
        cr.tl.lineages(adata_cflare)
        res = cr.tl.lineage_drivers(
            adata_cflare, return_drivers=True, use_raw=False, lineages=lineages
        )
        exp_lineages = adata_cflare.obsm["to_terminal_states"]
        if lineages is not None:
            exp_lineages = exp_lineages[lineages]
        exp_lineages = exp_lineages.names

        assert isinstance(res, pd.DataFrame)
        assert res.shape[0] == adata_cflare.n_vars

        for name in exp_lineages:
            assert np.all(adata_cflare.var[f"to {name} corr"] >= -1.0)
            assert np.all(adata_cflare.var[f"to {name} corr"] <= 1.0)
            # res is sorted
            np.testing.assert_allclose(
                adata_cflare.var[f"to {name} corr"],
                res[f"{name} corr"].loc[adata_cflare.var_names],
            )

            assert np.all(res[f"{name} pval"] >= 0)
            assert np.all(res[f"{name} pval"] <= 1.0)

            assert np.all(adata_cflare.var[f"to {name} qval"] >= 0)
            assert np.all(adata_cflare.var[f"to {name} qval"] <= 1.0)
            # res is sorted
            np.testing.assert_allclose(
                adata_cflare.var[f"to {name} qval"],
                res[f"{name} qval"].loc[adata_cflare.var_names],
            )

            assert np.all(res[f"{name} corr"] >= res[f"{name} ci low"])
            assert np.all(res[f"{name} corr"] <= res[f"{name} ci high"])

    def test_invalid_mode(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        with pytest.raises(ValueError):
            cr.tl.lineage_drivers(adata_cflare, use_raw=False, method="foobar")

    def test_invalid_n_perms_type(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        with pytest.raises(TypeError):
            cr.tl.lineage_drivers(
                adata_cflare, use_raw=False, n_perms=None, method="perm_test"
            )

    def test_invalid_n_perms_value(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        with pytest.raises(ValueError):
            cr.tl.lineage_drivers(
                adata_cflare, use_raw=False, n_perms=0, method="perm_test"
            )

    def test_seed_reproducible(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        res_a = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            n_perms=10,
            n_jobs=1,
            seed=0,
            method="perm_test",
        )
        res_b = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            n_perms=10,
            n_jobs=1,
            seed=0,
            method="perm_test",
        )
        res_diff_seed = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            n_perms=10,
            n_jobs=1,
            seed=1,
            method="perm_test",
        )

        assert res_a is not res_b
        np.testing.assert_array_equal(res_a.index, res_b.index)
        np.testing.assert_array_equal(res_a.columns, res_b.columns)
        np.testing.assert_allclose(res_a.values, res_b.values)

        assert not np.allclose(res_a.values, res_diff_seed.values)

    def test_seed_reproducible_parallel(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        res_a = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            n_perms=10,
            n_jobs=2,
            backend="threading",
            seed=42,
            method="perm_test",
        )
        res_b = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            n_perms=10,
            n_jobs=2,
            backend="threading",
            seed=42,
            method="perm_test",
        )

        assert res_a is not res_b
        np.testing.assert_array_equal(res_a.index, res_b.index)
        np.testing.assert_array_equal(res_a.columns, res_b.columns)
        np.testing.assert_allclose(res_a.values, res_b.values)

    def test_confidence_level(self, adata_cflare: AnnData):
        cr.tl.lineages(adata_cflare)
        res_narrow = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            confidence_level=0.95,
        )
        res_wide = cr.tl.lineage_drivers(
            adata_cflare,
            use_raw=False,
            return_drivers=True,
            confidence_level=0.99,
        )

        for name in ["0", "1"]:
            assert np.all(res_narrow[f"{name} ci low"] >= res_wide[f"{name} ci low"])
            assert np.all(res_narrow[f"{name} ci high"] <= res_wide[f"{name} ci high"])


class TestRootFinal:
    def test_find_root(self, adata: AnnData):
        cr.tl.initial_states(adata)

        assert str(TermStatesKey.BACKWARD) in adata.obs.keys()
        assert _probs(TermStatesKey.BACKWARD) in adata.obs.keys()

    def test_find_final(self, adata: AnnData):
        cr.tl.terminal_states(adata, n_states=5, fit_kwargs=dict(n_cells=5))

        assert str(TermStatesKey.FORWARD) in adata.obs.keys()
        assert _probs(TermStatesKey.FORWARD) in adata.obs.keys()

    def test_invalid_cluster_key(self, adata: AnnData):
        with pytest.raises(KeyError):
            cr.tl.initial_states(adata, cluster_key="foo")

    def test_invalid_weight(self, adata: AnnData):
        with pytest.raises(ValueError):
            cr.tl.initial_states(adata, weight_connectivities=10)

    def test_invalid_invalid_mode(self, adata: AnnData):
        with pytest.raises(ValueError):
            cr.tl.initial_states(adata, mode="foobar")

    @pytest.mark.parametrize("force_recompute", [False, True])
    def test_force_recompute(self, adata: AnnData, force_recompute: bool):
        cr.tl.terminal_states(adata, n_states=5, fit_kwargs=dict(n_cells=5))
        tmat1 = adata.obsp["T_fwd"]

        cr.tl.terminal_states(
            adata,
            n_states=5,
            fit_kwargs=dict(n_cells=5),
            force_recompute=force_recompute,
        )
        tmat2 = adata.obsp["T_fwd"]
        if force_recompute:
            assert tmat1 is not tmat2
            np.testing.assert_allclose(tmat1.A, tmat2.A)
            np.testing.assert_allclose(tmat1.sum(1), 1.0)
        else:
            assert tmat1 is tmat2


class TestTransitionMatrix:
    def test_invalid_velocity_key(self, adata: AnnData):
        with pytest.raises(KeyError):
            cr.tl.transition_matrix(adata, vkey="foo")

    def test_invalid_weight(self, adata: AnnData):
        with pytest.raises(ValueError):
            cr.tl.transition_matrix(adata, weight_connectivities=-1)

    def test_invalid_conn_key(self, adata: AnnData):
        with pytest.raises(KeyError):
            cr.tl.transition_matrix(adata, conn_key="foo")

    def test_forward(self, adata: AnnData):
        kernel_add = cr.tl.transition_matrix(adata, backward=False, softmax_scale=None)

        assert isinstance(kernel_add, KernelAdd)
        assert not kernel_add.backward

    def test_only_connectivities(self, adata: AnnData):
        ck = cr.tl.transition_matrix(adata, weight_connectivities=1)

        assert isinstance(ck, cr.tl.kernels.ConnectivityKernel)

    def test_connectivities_conn_key(self, adata: AnnData):
        key = "foobar"
        assert key not in adata.obsp
        adata.obsp[key] = np.eye(adata.n_obs)

        ck = cr.tl.transition_matrix(adata, weight_connectivities=1, conn_key=key)

        assert isinstance(ck, cr.tl.kernels.ConnectivityKernel)

        np.testing.assert_array_equal(ck.transition_matrix, adata.obsp[key])

    def test_only_velocity(self, adata: AnnData):
        vk = cr.tl.transition_matrix(adata, weight_connectivities=0)

        assert isinstance(vk, cr.tl.kernels.VelocityKernel)

    def test_backward(self, adata: AnnData):
        kernel_add = cr.tl.transition_matrix(adata, backward=True, softmax_scale=None)

        assert isinstance(kernel_add, KernelAdd)
        assert kernel_add.backward



from typing import Any, Union, Optional

from typing_extensions import Literal

from anndata import AnnData
from cellrank.ul._docs import d
from cellrank.tl.estimators import GPCCA
from cellrank.tl.kernels._precomputed_kernel import DummyKernel

import numpy as np
import pandas as pd

_error = None
try:
    from statot import OTKernel as OTKernel_
except ImportError as e:
    from cellrank.external.kernels._import_error_kernel import (
        ErroredKernel as OTKernel_,
    )

    _error = e


@d.dedent
class StationaryOTKernel(OTKernel_, error=_error):
    """
    Stationary optimal transport kernel from :cite:`zhang:21`.

    This class requires the `statOT` package, which can be installed as `pip install statot POT`.

    Parameters
    ----------
    %(adata)s
    terminal_states
        Categorical :class:`pandas.Series` where non-`NaN` values mark terminal states.
        If `None`, terminal states are assumed to be already present in :attr:`adata` ``['terminal_states']``.
    g
        Key in :attr:`anndata.AnnData.obs` containing relative growth rates for cells or the array itself.
    cluster_key
        If a key to cluster labels is given, `terminal_states` will be associated with these for naming and colors.
    kwargs
        Additional keyword arguments.
    """

    __import_error_message__ = "Unable to import the kernel. Please install `statOT` first as `pip install statot POT`."

    def __init__(
        self,
        adata: AnnData,
        g: Union[str, np.ndarray],
        terminal_states: Optional[Union[str, pd.Series]] = None,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ):
        if terminal_states is not None:
            dk = DummyKernel(adata, backward=False)
            estim = GPCCA(dk, write_to_adata=True)
            estim.set_terminal_states(terminal_states, cluster_key=cluster_key)

        try:
            super().__init__(adata, g=g, **kwargs)
        except Exception as e:  # noqa: B902
            raise RuntimeError("Unable to initialize the kernel.") from e

    def _read_from_adata(
        self,
        conn_key: Optional[str] = "connectivities",
        read_conn: bool = True,
        **kwargs: Any,
    ) -> None:
        super()._read_from_adata(conn_key=conn_key, read_conn=False, **kwargs)

    def compute_transition_matrix(
        self,
        eps: float,
        dt: float,
        basis: str = "X_pca",
        cost_norm_method: Optional[str] = None,
        method: Literal["ent", "quad", "unbal"] = "ent",
        tol: float = 0.0,
        thresh: float = 0.0,
        maxiter: int = 5000,
        C: Optional[np.ndarray] = None,
        verbose: bool = False,
        **kwargs: Any,
    ) -> "StationaryOTKernel":
        """
        Compute transition matrix using stationary OT :cite:`zhang:21`.

        Parameters
        ----------
        eps
            Regularization parameter.
        dt
            Choice of the time step over which to fit the model.
        basis
            Key in :attr:`anndata.AnnData.obsm` where the basis is stored.
        cost_norm_method
            Cost normalization method to use. Use "mean" to ensure `mean(C) = 1` or refer to
            :func:`ot.utils.cost_normalization` for more information.
        method
            Choice of regularization. Valid options are:

                - `'ent'`: entropy.
                - `'quad'`: L2-norm.
                - `'unbal'`: unbalanced transport (not yet implemented).

        tol
            Relative tolerance for OT solver convergence.
        thresh
            Threshold for output transition probabilities.
        maxiter
            Maximum number of iterations for OT solver.
        C
            Cost matrix for optimal transport problem.
        verbose
            Detailed output on convergence of OT solver.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        :class:`cellrank.external.kernels.StationaryOTKernel`
            Makes :attr:`transition_matrix` available.
        """
        if method not in ("ent", "quad", "unbal"):
            raise ValueError(
                f"Invalid method `{method!r}`. Valid options are: `{['ent', 'quad', 'unbal']}`."
            )
        if method == "unbal":
            raise NotImplementedError("Method `'unbal'` is not yet implemented.")

        super().compute_transition_matrix(
            eps,
            dt,
            expr_key=basis,
            cost_norm_method=cost_norm_method,
            method=method,
            tol=tol,
            thresh=thresh,
            maxiter=maxiter,
            C=C,
            verbose=verbose,
            **kwargs,
        )

        return self



"""Module for documentation helper function."""

from textwrap import dedent

from docrep import DocstringProcessor

_adata = """\
adata : :class:`anndata.AnnData`
    Annotated data object."""
_adata_ret = """\
:class:`anndata.AnnData`
    Annotated data object."""
_plotting = """\
figsize
    Size of the figure.
dpi
    Dots per inch.
save
    Filename where to save the plot."""
_n_jobs = """\
n_jobs
    Number of parallel jobs. If `-1`, use all available cores. If `None` or `1`, the execution is sequential."""
_parallel = f"""\
show_progress_bar
    Whether to show a progress bar. Disabling it may slightly improve performance.
{_n_jobs}
backend
    Which backend to use for parallelization. See :class:`joblib.Parallel` for valid options."""
_model = """\
model
    Model based on :class:`cellrank.ul.models.BaseModel` to fit.

    If a :class:`dict`, gene and lineage specific models can be specified. Use ``'*'`` to indicate
    all genes or lineages, for example ``{'gene_1': {'*': ...}, 'gene_2': {'lineage_1': ..., '*': ...}}``."""
_just_plots = """\
None
    Nothing, just plots the figure. Optionally saves it based on ``save``."""
_plots_or_returns_models = """\
None
    If ``return_models=False``, just plots the figure and optionally saves it based on ``save``.
Dict[str, Dict[str, :class:`cellrank.ul.models.BaseModel`]]
    Otherwise returns the fitted models as ``{'gene_1': {'lineage_1': <model_11>, ...}, ...}``.
    Models which have failed will be instances of :class:`cellrank.ul.models.FailedModel`."""
_backward = """\
backward
    Direction of the process."""
_eigen = """\
which
    Eigenvalues are in general complex. `'LR'` - largest real part, `'LM'` - largest magnitude.
alpha
    Used to compute the `eigengap`. ``alpha`` is the weight given to the deviation of an eigenvalue from one."""
_n_cells = """\
n_cells
    Number of most likely cells from each macrostate to select."""
_fit = """\
n_lineages
    Number of lineages. If `None`, it will be determined automatically.
cluster_key
    Match computed states against pre-computed clusters to annotate the states.
    For this, provide a key from :attr:`adata` ``.obs`` where cluster labels have been computed.
keys
    Determines which %(initial_or_terminal)s states to use by passing their names.
    Further, %(initial_or_terminal)s states can be combined. If e.g. the %(terminal)s states are
    ['Neuronal_1', 'Neuronal_1', 'Astrocytes', 'OPC'], then passing ``keys=['Neuronal_1, Neuronal_2', 'OPC']``
    means that the two neuronal %(terminal)s states are treated as one and the 'Astrocyte' state is excluded."""
_density_correction = (
    "Optionally, we apply a density correction as described in :cite:`coifman:05`, "
    "where we use the implementation of :cite:`haghverdi:16`."
)
_time_range = """\
time_range
    Specify start and end times:

        - If a :class:`tuple`, it specifies the minimum and maximum pseudotime. Both values can be `None`,
          in which case the minimum is the earliest pseudotime and the maximum is automatically determined.
        - If a :class:`float`, it specifies the maximum pseudotime."""

_velocity_mode = """\
mode
    How to compute transition probabilities. Valid options are:

        - `{m.DETERMINISTIC.s!r}` - deterministic computation that doesn't propagate uncertainty.
        - `{m.MONTE_CARLO.s!r}` - Monte Carlo average of randomly sampled velocity vectors.
        - `{m.STOCHASTIC.s!r}` - second order approximation, only available when :mod:`jax` is installed.
        - `{m.SAMPLING.s!r}` - sample 1 transition matrix from the velocity distribution."""
_velocity_backward_mode = """\
backward_mode
    Only matters if initialized as :attr:`backward` ``=True``.  Valid options are:

        - `{b.TRANSPOSE.s!r}` - compute transitions from neighboring cells `j` to cell `i`.
        - `{b.NEGATE.s!r}` - negate the velocity vector."""
_velocity_backward_mode_high_lvl = """\
backward_mode
    How to compute the backward transitions. Valid options are:

        - `{b.TRANSPOSE.s!r}` - compute transitions from neighboring cells `j` to cell `i`.
        - `{b.NEGATE.s!r}` - negate the velocity vector."""
_copy = """Return a copy of self."""
_initial = "initial"
_terminal = "terminal"
_model_callback = """\
callback
    Function which takes a :class:`cellrank.ul.models.BaseModel` and some keyword arguments
    for :meth:`cellrank.ul.models.BaseModel.prepare` and returns the prepared model.
    Can be specified in gene- and lineage-specific manner, similarly to :attr:`model`."""
_genes = """\
genes
    Genes in ``adata.var_names`` or in ``adata.raw.var_names``, if ``use_raw=True``."""
_softmax_scale = """\
softmax_scale
    Scaling parameter for the softmax. If `None`, it will be estimated using ``1 / median(correlations)``.
    The idea behind this is to scale the softmax to counteract everything tending to orthogonality in high dimensions."""  # noqa
_time_mode = """\
mode
    Can be either `'embedding'` or `'time'`:

        - `'embedding'` - plot the embedding while coloring in the absorption probabilities.
        - `'time'` - plot the pseudotime on x-axis and the absorption probabilities on y-axis."""
_write_to_adata = """\
Updates the :attr:`adata` with the following fields:

        - ``.obsp['{{key}}']`` - the transition matrix.
        - ``.uns['{{key}}_params']`` - parameters used for calculation."""
_en_cutoff_p_thresh = """\
en_cutoff
    If ``cluster_key`` is given, this parameter determines when an approximate recurrent class will
    be labeled as *'Unknown'*, based on the entropy of the distribution of cells over transcriptomic clusters.
p_thresh
    If cell cycle scores were provided, a *Wilcoxon rank-sum test* is conducted to identify cell-cycle states.
    If the test returns a positive statistic and a p-value smaller than ``p_thresh``, a warning will be issued."""
_return_models = """\
return_models
    If `True`, return the fitted models for each gene in ``genes`` and lineage in ``lineages``."""
_basis = """\
basis
    Basis to use when ``mode='embedding'``. If `None`, use `'umap'`."""
_velocity_scheme = """\
scheme
    Similarity scheme between cells as described in :cite:`li:20`. Can be one of the following:

        - `{s.DOT_PRODUCT.s!r}`: :class:`cellrank.tl.kernels.DotProductScheme`.
        - `{s.COSINE.s!r}`: :class:`cellrank.tl.kernels.CosineScheme`.
        - `{s.CORRELATION.s!r}`: :class:`cellrank.tl.kernels.CorrelationScheme`.

    Alternatively, any function can be passed as long as it follows the signature of
    :meth:`cellrank.tl.kernels.SimilaritySchemeABC.__call__`."""
_cond_num = """\
compute_cond_num
    Whether to compute condition number of the transition matrix. Note that this might be costly,
    since it does not use sparse implementation."""
_soft_scheme_fmt = """\
b
    The growth rate of generalized logistic function.{}
nu
    Affects near which asymptote maximum growth occurs.{}"""
_rw_ixs = """\
Can be specified as either a :class:`dict` with a key corresponding to cluster key in
    :attr:`anndata.AnnData.obs` and values to clusters or just a sequence of cell ids
    in :attr:`anndata.AnnData.obs_names`."""


def inject_docs(**kwargs):  # noqa
    def decorator(obj):
        obj.__doc__ = dedent(obj.__doc__).format(**kwargs)
        return obj

    def decorator2(obj):
        obj.__doc__ = dedent(kwargs["__doc__"])
        return obj

    if isinstance(kwargs.get("__doc__", None), str) and len(kwargs) == 1:
        return decorator2

    return decorator


d = DocstringProcessor(
    plotting=_plotting,
    n_jobs=_n_jobs,
    parallel=_parallel,
    model=_model,
    adata=_adata,
    adata_ret=_adata_ret,
    just_plots=_just_plots,
    backward=_backward,
    initial=_initial,
    terminal=_terminal,
    eigen=_eigen,
    initial_or_terminal=f"{_initial} or {_terminal}",
    n_cells=_n_cells,
    fit=_fit,
    copy=_copy,
    density_correction=_density_correction,
    time_range=_time_range,
    velocity_mode=_velocity_mode,
    velocity_backward_mode=_velocity_backward_mode,
    velocity_backward_mode_high_lvl=_velocity_backward_mode_high_lvl,
    model_callback=_model_callback,
    genes=_genes,
    softmax_scale=_softmax_scale,
    time_mode=_time_mode,
    write_to_adata=_write_to_adata,
    en_cutoff_p_thresh=_en_cutoff_p_thresh,
    return_models=_return_models,
    plots_or_returns_models=_plots_or_returns_models,
    basis=_basis,
    velocity_scheme=_velocity_scheme,
    cond_num=_cond_num,
    soft_scheme=_soft_scheme_fmt.format("", "", ""),
    soft_scheme_kernel=_soft_scheme_fmt.format(
        *([" Only used when `threshold_scheme='soft'`."] * 3)
    ),
    rw_ixs=_rw_ixs,
)



"""General utility functions module."""
import pickle
from types import MappingProxyType
from typing import Any, Dict, List, Tuple, Union, TypeVar, Iterable, Optional
from pathlib import Path
from functools import wraps, update_wrapper
from multiprocessing import cpu_count

from cellrank import logging as logg
from cellrank.ul._docs import d

import numpy as np
from scipy.sparse import issparse, spmatrix

AnnData = TypeVar("AnnData")


class Pickleable:
    """Class which allows serialization and deserialization using :mod:pickle."""

    @d.get_full_description(base="pickleable")
    @d.get_sections(base="pickleable", sections=["Parameters", "Returns"])
    def write(self, fname: Union[str, Path], ext: Optional[str] = "pickle") -> None:
        """
        Serialize self to a file.

        Parameters
        ----------
        fname
            Filename where to save the object.
        ext
            Filename extension to use. If `None`, don't append any extension.

        Returns
        -------
        None
            Nothing, just writes itself to a file using :mod:`pickle`.
        """

        fname = str(fname)
        if ext is not None:
            if not ext.startswith("."):
                ext = "." + ext
            if not fname.endswith(ext):
                fname += ext

        logg.debug(f"Writing to `{fname}`")

        with open(fname, "wb") as fout:
            pickle.dump(self, fout)

    @staticmethod
    def read(fname: Union[str, Path]) -> Any:
        """
        Deserialize self from a file.

        Parameters
        ----------
        fname
            Filename from which to read the object.

        Returns
        -------
        :class:`typing.Any`
            The deserialized object.
        """

        with open(fname, "rb") as fin:
            return pickle.load(fin)


def _check_collection(
    adata: AnnData,
    needles: Iterable[str],
    attr_name: str,
    key_name: str = "Gene",
    use_raw: bool = False,
    raise_exc: bool = True,
) -> List[str]:
    """
    Check if given collection contains all the keys.

    Parameters
    ----------
    adata: :class:`anndata.AnnData`
        Annotated data object.
    needles
        Keys to check.
    attr_name
        Attribute of ``adata`` where the needles are stored.
    key_name
        Pretty name of the key which will be displayed when error is found.
    use_raw
        Whether to access ``adata.raw`` or just ``adata``.

    Returns
    -------
    None
        Nothing, but raises and :class:`KeyError` if one of the needles is not found.
    """
    adata_name = "adata"

    if use_raw and adata.raw is None:
        logg.warning(
            "Argument `use_raw` was set to `True`, but no `raw` attribute is found. Ignoring"
        )
        use_raw = False
    if use_raw:
        adata_name = "adata.raw"
        adata = adata.raw

    haystack, res = getattr(adata, attr_name), []
    for needle in needles:
        if needle not in haystack:
            if raise_exc:
                raise KeyError(
                    f"{key_name} `{needle}` not found in `{adata_name}.{attr_name}`."
                )
        else:
            res.append(needle)

    return res


def _get_n_cores(n_cores: Optional[int], n_jobs: Optional[int]) -> int:
    """
    Make number of cores a positive integer.

    Parameters
    ----------
    n_cores
        Number of cores to use.
    n_jobs.
        Number of jobs. This is just used to determine if the collection is a singleton.
        If `1`, always returns `1`.

    Returns
    -------
    int
        Positive integer corresponding to how many cores to use.
    """
    if n_cores == 0:
        raise ValueError("Number of cores cannot be `0`.")
    if n_jobs == 1 or n_cores is None:
        return 1
    if n_cores < 0:
        return cpu_count() + 1 + n_cores

    return n_cores


def _minmax(
    data: np.ndarray, perc: Optional[Tuple[float, float]] = None
) -> Tuple[float, float]:
    """
    Return minimum and maximum value of the data.

    Parameters
    ----------
    data
        Values for which to return the minimum and maximum.
    perc
        If not `None`, clip the values by the percentiles before getting the result.

    Returns
    -------
    :class:`tuple`
        Minimum and maximum values, respectively.
    """
    if perc is not None:
        data = np.clip(data, *np.percentile(data, sorted(perc)))

    return float(np.nanmin(data)), float(np.nanmax(data))


def _modify_neigh_key(key: Optional[str]) -> str:
    if key in (None, "connectivities", "distances"):
        return "neighbors"

    if key.endswith("_connectivities"):
        key = key[:-15]
    elif key.endswith("_distances"):
        key = key[:-10]

    return key


def _has_neighs(adata: AnnData, key: Optional[str] = None) -> bool:
    return _modify_neigh_key(key) in adata.uns.keys()


def _get_neighs(
    adata: AnnData, mode: str = "distances", key: Optional[str] = None
) -> Union[np.ndarray, spmatrix]:
    if key is None:
        res = _read_graph_data(adata, mode)  # legacy behavior
    else:
        try:
            res = _read_graph_data(adata, key)
            assert isinstance(res, (np.ndarray, spmatrix))
        except (KeyError, AssertionError):
            res = _read_graph_data(adata, f"{_modify_neigh_key(key)}_{mode}")

    if not isinstance(res, (np.ndarray, spmatrix)):
        raise TypeError(
            f"Expected to find `numpy.ndarray` or `scipy.sparse.spmatrix`, found `{type(res)}`."
        )

    return res


def _get_neighs_params(adata: AnnData, key: str = "neighbors") -> Dict[str, Any]:
    return adata.uns.get(key, {}).get("params")


def _read_graph_data(adata: AnnData, key: str) -> Union[np.ndarray, spmatrix]:
    """
    Read graph data from :mod:`anndata`.

    Parameters
    ----------
    adata
        Annotated data object.
    key
        Key in ``adata.obsp``.

    Returns
    -------
    :class:`numpy.ndarray` or :class:`scipy.sparse.spmatrix`
        The graph data.
    """

    logg.debug(f"Reading key `{key!r}` from `adata.obsp`")
    if key in adata.obsp.keys():
        return adata.obsp[key]

    raise KeyError(f"Unable to find key `{key!r}` in `adata.obsp`.")


def _write_graph_data(
    adata: AnnData,
    data: Union[np.ndarray, spmatrix],
    key: str,
):
    """
    Write graph data to :mod:`AnnData`.

    :module`anndata` >=0.7 stores `(n_obs x n_obs)` matrices in `.obsp` rather than `.uns`.
    This is for backward compatibility.

    Parameters
    ----------
    adata
        Annotated data object.
    data
        The graph data we want to write.
    key
        Key from either ``adata.uns`` or `adata.obsp``.

    Returns
    --------
    None
        Nothing, just writes the data.
    """

    try:
        adata.obsp[key] = data
        write_to = "obsp"

        if data.shape[0] != data.shape[1]:
            logg.warning(
                f"`adata.obsp` attribute should only contain square matrices, found shape `{data.shape}`"
            )

    except AttributeError:
        adata.uns[key] = data
        write_to = "uns"

    logg.debug(f"Writing graph data to `adata.{write_to}[{key!r}]`")


def valuedispatch(func):
    """Dispatch a function based on the first value."""
    registry = {}

    def dispatch(value):
        return registry.get(value, func)

    def register(value, func=None):
        if func is None:
            return lambda f: register(value, f)
        registry[value] = func
        return func

    @wraps(func)
    def wrapper(*args, **kwargs):
        return dispatch(args[0])(*args[1:], **kwargs)

    wrapper.register = register
    wrapper.dispatch = dispatch
    wrapper.registry = MappingProxyType(registry)

    update_wrapper(wrapper, func)

    return wrapper


def _densify_squeeze(x: Union[spmatrix, np.ndarray], dtype=np.float32) -> np.ndarray:
    if issparse(x):
        x = x.toarray()
    # use np.array instead of asarray to create a copy
    x = np.array(x, dtype=dtype)
    if x.ndim == 2 and x.shape[1] == 1:
        x = np.squeeze(x, axis=1)

    return x



"""Module containing CellRank constants."""

from abc import ABC, ABCMeta
from enum import Enum, EnumMeta
from typing import Any, Union, Callable
from functools import wraps

from cellrank.ul._docs import _initial, _terminal

_DEFAULT_BACKEND = "loky"


class PrettyEnum(Enum):
    """Enum with a pretty __str__ and __repr__."""

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(self.value)

    @property
    def v(self) -> Any:
        """Return the value."""
        return self.value

    @property
    def s(self) -> str:
        """Return the value as string."""
        return str(self)

    @property
    def fmt(self) -> Callable:
        """Return the formatting function given by the string value."""
        return self.s.format


class Lin(PrettyEnum):
    """Lineage aggregation type."""

    REST = "rest"
    OTHERS = "others"


class Direction(PrettyEnum):
    """Direction of the process."""

    FORWARD = "fwd"
    BACKWARD = "bwd"


class DirectionPlot(PrettyEnum):
    """Pretty direction names for plotting."""

    FORWARD = "forward"
    BACKWARD = "backward"


class DirPrefix(PrettyEnum):
    """Direction prefix."""

    FORWARD = "to"
    BACKWARD = "from"


class TermStatesKey(PrettyEnum):
    """State key in `adata.obs`."""

    FORWARD = f"{_terminal}_states"
    BACKWARD = f"{_initial}_states"


class TerminalStatesPlot(PrettyEnum):
    """Pretty state names for pl."""

    FORWARD = TermStatesKey.FORWARD.s.replace("_", " ")
    BACKWARD = TermStatesKey.BACKWARD.s.replace("_", " ")


class Macro(PrettyEnum):
    """Key for scaled, subsampled macrostates in `adata.obsm`."""

    FORWARD = f"macrostates_{Direction.FORWARD}"
    BACKWARD = f"macrostates_{Direction.BACKWARD}"


# TermStatesKey and AbsProbKey must have the same suffix `_..._states` because of model.prepare
class AbsProbKey(PrettyEnum):
    """Lineage key in `adata.obsm`."""

    FORWARD = f"{DirPrefix.FORWARD}_{TermStatesKey.FORWARD}"
    BACKWARD = f"{DirPrefix.BACKWARD}_{TermStatesKey.BACKWARD}"


class ThresholdScheme(PrettyEnum):
    """Pseudotime thresholding scheme."""

    SOFT = "soft"
    HARD = "hard"


def _transition(d: Union[str, Direction]) -> str:
    return f"T_{d}"


def _lin_names(k: Union[str, AbsProbKey]) -> str:
    return f"{k}_names"


def _colors(k: Union[str, AbsProbKey, TermStatesKey]) -> str:
    return f"{k}_colors"


def _probs(k: Union[str, TermStatesKey]) -> str:
    return f"{k}_probs"


def _pd(k: Union[str, AbsProbKey]) -> str:
    return f"{k}_priming_degree"


def _pretty_raise_enum(cls, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except ValueError as e:
            _cls, value, *_ = args
            e.args = (cls._format(value),)
            raise e

    if not issubclass(cls, ErrorFormatterABC):
        raise TypeError(f"Class `{cls}` must be subtype of `ErrorFormatterABC`.")
    elif not len(cls.__members__):
        # empty enum, for class hierarchy
        return fun

    return wrapper


class ABCEnumMeta(EnumMeta, ABCMeta):  # noqa
    def __call__(cls, *args, **kw):  # noqa
        if getattr(cls, "__error_format__", None) is None:
            raise TypeError(
                f"Can't instantiate class `{cls.__name__}` "
                f"without `__error_format__` class attribute."
            )
        return super().__call__(*args, **kw)

    def __new__(cls, clsname, superclasses, attributedict):  # noqa
        res = super().__new__(cls, clsname, superclasses, attributedict)
        res.__new__ = _pretty_raise_enum(res, res.__new__)
        return res


class ErrorFormatterABC(ABC):  # noqa
    __error_format__ = "Invalid option `{!r}` for `{}`. Valid options are: `{}`."

    @classmethod
    def _format(cls, value):
        return cls.__error_format__.format(
            value, cls.__name__, [m.value for m in cls.__members__.values()]
        )


class ModeEnum(ErrorFormatterABC, PrettyEnum, metaclass=ABCEnumMeta):  # noqa
    pass



import os
from sys import version_info
from typing import Tuple, Optional
from pathlib import Path

from filelock import FileLock

from cellrank.ul.models import GAM, GAMR, SKLearnModel

os.environ["NUMBA_NUM_THREADS"] = "2"

import pytest
from _helpers import create_model

import scanpy as sc
import cellrank as cr
from anndata import AnnData
from cellrank.tl.kernels import VelocityKernel, ConnectivityKernel
from cellrank.tl._constants import AbsProbKey
from cellrank.tl.estimators import GPCCA, CFLARE

import numpy as np

import matplotlib

matplotlib.use("Agg")
np.random.seed(42)

_adata_small = sc.read("tests/_ground_truth_adatas/adata_50.h5ad")
_adata_medium = sc.read("tests/_ground_truth_adatas/adata_100.h5ad")
_adata_large = sc.read("tests/_ground_truth_adatas/adata_200.h5ad")


def _create_cflare(*, backward: bool = False) -> Tuple[AnnData, CFLARE]:
    adata = _adata_medium.copy()

    sc.tl.paga(adata, groups="clusters")

    vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
        softmax_scale=4
    )
    ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix()
    final_kernel = 0.8 * vk + 0.2 * ck

    mc = CFLARE(final_kernel)

    mc.compute_partition()
    mc.compute_eigendecomposition()
    mc.compute_terminal_states(use=2)
    mc.compute_absorption_probabilities(use_petsc=False)
    mc.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

    assert adata is mc.adata
    if backward:
        assert str(AbsProbKey.BACKWARD) in adata.obsm
    else:
        assert str(AbsProbKey.FORWARD) in adata.obsm
    np.testing.assert_allclose(mc.absorption_probabilities.X.sum(1), 1.0, rtol=1e-6)

    return adata, mc


def _create_gpcca(*, backward: bool = False) -> Tuple[AnnData, GPCCA]:
    adata = _adata_medium.copy()

    sc.tl.paga(adata, groups="clusters")

    vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
        softmax_scale=4
    )
    ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix()
    final_kernel = 0.8 * vk + 0.2 * ck

    mc = GPCCA(final_kernel)

    mc.compute_partition()
    mc.compute_eigendecomposition()
    mc.compute_schur(method="krylov")
    mc.compute_macrostates(n_states=2)
    mc.set_terminal_states_from_macrostates()
    mc.compute_absorption_probabilities()
    mc.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

    assert adata is mc.adata
    if backward:
        assert str(AbsProbKey.BACKWARD) in adata.obsm
    else:
        assert str(AbsProbKey.FORWARD) in adata.obsm
    np.testing.assert_allclose(mc.absorption_probabilities.X.sum(1), 1.0, rtol=1e-6)

    return adata, mc


def _create_gamr_model(_adata: AnnData) -> Optional[GAMR]:
    try:
        m = GAMR(_adata)
        m.prepare(_adata.var_names[0], "0").fit()
        m.predict(level=0.95)
        return m
    except:
        return None


@pytest.fixture
def adata() -> AnnData:
    return _adata_small.copy()


@pytest.fixture
def adata_large() -> AnnData:
    return _adata_large.copy()


@pytest.fixture
def adata_cflare_fwd(
    adata_cflare=_create_cflare(backward=False),
) -> Tuple[AnnData, CFLARE]:
    adata, cflare = adata_cflare
    return adata.copy(), cflare


@pytest.fixture
def adata_gpcca_fwd(adata_gpcca=_create_gpcca(backward=False)) -> Tuple[AnnData, GPCCA]:
    adata, gpcca = adata_gpcca
    return adata.copy(), gpcca


@pytest.fixture
def adata_gpcca_bwd(adata_gpcca=_create_gpcca(backward=True)) -> Tuple[AnnData, GPCCA]:
    adata, gpcca = adata_gpcca
    return adata.copy(), gpcca


@pytest.fixture
def adata_cflare(adata_cflare=_create_cflare(backward=False)) -> AnnData:
    return adata_cflare[0].copy()


@pytest.fixture(scope="session")
def adata_gamr(adata_cflare=_create_cflare(backward=False)) -> AnnData:
    return adata_cflare[0].copy()


@pytest.fixture(scope="session")
def gamr_model(
    adata_gamr: AnnData, tmp_path_factory: Path, worker_id: str
) -> Optional[GAMR]:
    model = None

    if version_info[:2] <= (3, 6):
        pytest.skip("Pickling of Enums doesn't work in Python3.6.")
    elif worker_id == "master":
        model = _create_gamr_model(adata_gamr)
    else:
        root_tmp_dir = tmp_path_factory.getbasetemp().parent
        fn = root_tmp_dir / "model.pickle"

        with FileLock(f"{fn}.lock"):
            if fn.is_file():
                model = GAMR.read(fn)
            else:
                model = _create_gamr_model(adata_gamr)
                if model is not None:
                    model.write(fn)

    if model is None:
        pytest.skip("Unable to create `cellrank.ul.models.GAMR`.")

    return model


@pytest.fixture
def pygam_model(adata_cflare: AnnData) -> GAM:
    m = GAM(adata_cflare)
    m.prepare(adata_cflare.var_names[0], "0").fit()
    m.predict()
    m.confidence_interval()

    return m


@pytest.fixture
def sklearn_model(adata_cflare: AnnData) -> SKLearnModel:
    m = create_model(adata_cflare)
    assert isinstance(m, SKLearnModel), m

    m.prepare(adata_cflare.var_names[0], "0").fit()
    m.predict()
    m.confidence_interval()

    return m


@pytest.fixture
def lineage():
    x = cr.tl.Lineage(
        np.array(
            [
                [1.23459664e-01, 1.29965675e-01, 1.92828002e-01, 9.39402664e-01],
                [1.05635239e00, 4.45833459e-01, 2.29080759e00, 1.90132652e00],
                [6.77880737e-02, 4.97556864e-02, 1.18428661e00, 2.02318999e-01],
                [4.87500398e-01, 1.00657498e00, 2.20834882e-02, 5.03008905e-01],
                [6.27190917e00, 7.27864781e00, 1.03978903e00, 1.55903460e01],
                [3.85149269e-01, 3.54765380e-01, 1.77871487e-01, 8.22138648e-02],
                [7.06618729e00, 1.33133671e01, 1.44904591e00, 5.79813391e00],
                [8.18005744e-02, 5.36844933e-01, 1.86646162e00, 2.41141727e00],
                [1.44892035e-01, 2.34036215e-01, 6.32392890e-01, 1.13211403e-02],
                [2.44926466e-01, 2.50293183e-01, 1.77540208e-01, 3.27240144e-01],
            ]
        ),
        names=["foo", "bar", "baz", "quux"],
    )
    return x / x.sum(1)


@pytest.fixture
def kernel(adata_large: AnnData):
    vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
    ck = ConnectivityKernel(adata_large).compute_transition_matrix()
    return (0.8 * vk + 0.2 * ck).compute_transition_matrix()


# removes overly verbose logging errors for rpy2
# see: https://github.com/pytest-dev/pytest/issues/5502#issuecomment-647157873
def pytest_sessionfinish(session, exitstatus):
    import logging

    loggers = [logging.getLogger()] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            logger.removeHandler(handler)


@pytest.fixture(scope="session")
def test_matrix_1() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - connected
        - irreducible
        - not reversible
    """

    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.4, 0.0, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0],  # 11
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_2() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - connected
        - not irreducible (1 recurrent, 1 transient class)
        - not reversible
    """
    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.  12.  13.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.3, 0.0, 0.0, 0.1, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2, 0.0, 0.0],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0],  # 11

        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.4],  # 12
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.7],  # 13
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_3() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - not connected
        - not irreducible (2 recurrent classes)
        - not reversible
    """

    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.  12.  13.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.4, 0.0, 0.0, 0.0, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2, 0.0, 0.0],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0],  # 11

        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.4],  # 12
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.7],  # 13
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_4() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Symmetric matrix. Not a transition matrix.
    """

    # fmt: off
    p = np.array(
        [
            # 0.   1.   2.   3.
            [0.0, 0.8, 0.2, 0.0],  # 0
            [0.8, 0.0, 0.6, 0.2],  # 1
            [0.2, 0.6, 0.0, 0.2],  # 2
            [0.0, 0.2, 0.2, 0.0],  # 3
        ]
    )
    # fmt: on

    return p



from typing import Any, List, Union, Optional, Sequence
from itertools import chain

from cellrank import logging as logg
from cellrank.ul._docs import d
from cellrank.ul._parallelize import parallelize

import numpy as np
from scipy.sparse import issparse, spmatrix


class RandomWalk:
    """
    Class that simulates a random walk on a Markov chain.

    Parameters
    ----------
    transition_matrix
        Row-stochastic transition matrix.
    start_ixs
        Indices from which to uniformly sample the starting points. If `None`, use all points.
    stop_ixs
        Indices which when hit, the random walk is terminated.
    """

    def __init__(
        self,
        transition_matrix: Union[np.ndarray, spmatrix],
        start_ixs: Optional[Sequence[int]] = None,
        stop_ixs: Optional[Sequence[int]] = None,
    ):
        if transition_matrix.ndim != 2 or (
            transition_matrix.shape[0] != transition_matrix.shape[1]
        ):
            raise ValueError("Array is not a square matrix.")
        if not np.allclose(transition_matrix.sum(1), 1.0):
            raise ValueError("Transition matrix is not row-stochastic.")

        self._tmat = transition_matrix
        self._ixs = np.arange(self._tmat.shape[0])
        self._is_sparse = issparse(self._tmat)
        self._stop_ixs = set([] if stop_ixs is None or not len(stop_ixs) else stop_ixs)
        self._starting_dist = (
            np.ones_like(self._ixs)
            if start_ixs is None
            else np.isin(self._ixs, start_ixs)
        )
        if np.sum(self._starting_dist) == 0:
            raise ValueError("No starting indices have been selected.")

        self._starting_dist = self._starting_dist.astype(np.float64) / np.sum(
            self._starting_dist
        )

    def _should_stop(self, ix: int) -> bool:
        return ix in self._stop_ixs

    def _sample(self, ix: int, *, rs: np.random.RandomState) -> int:
        return rs.choice(
            self._ixs,
            p=self._tmat[ix].A.squeeze() if self._is_sparse else self._tmat[ix],
        )

    def _max_iter(self, max_iter: Union[int, float]) -> int:
        max_iter = (
            int(np.ceil(max_iter * len(self._ixs)))
            if isinstance(max_iter, float)
            else max_iter
        )
        if max_iter <= 1:
            raise ValueError(
                f"Expected number of iterations to be > 1, found `{max_iter}`."
            )
        return max_iter

    @d.get_sections(base="rw_sim", sections=["Parameters"])
    def simulate_one(
        self,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
    ) -> np.ndarray:
        """
        Simulate one random walk.

        Parameters
        ----------
        max_iter
            Maximum number of steps of a random walk. If a :class:`float`, it can be specified
            as a fraction of the number of cells.
        seed
            Random seed.
        successive_hits
            Number of successive hits in the ``stop_ixs`` required to stop prematurely.

        Returns
        -------
        Array of shape ``(max_iter + 1,)`` of states that have been visited. If ``stop_ixs`` was specified, the array
        may have smaller shape.
        """
        max_iter = self._max_iter(max_iter)
        if successive_hits < 0:
            raise ValueError(
                f"Expected number of successive hits to be positive, found `{successive_hits}`."
            )

        rs = np.random.RandomState(seed)
        ix = rs.choice(self._ixs, p=self._starting_dist)
        sim, cnt = [ix], -1

        for _ in range(max_iter):
            ix = self._sample(ix, rs=rs)
            sim.append(ix)
            cnt = (cnt + 1) if self._should_stop(ix) else -1
            if cnt >= successive_hits:
                break

        return np.array(sim)

    def _simulate_many(
        self,
        sims: np.ndarray,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
        queue: Optional[Any] = None,
    ) -> List[np.ndarray]:
        res = []
        for s in sims:
            sim = self.simulate_one(
                max_iter=max_iter,
                seed=None if seed is None else seed + s,
                successive_hits=successive_hits,
            )
            res.append(sim)
            if queue is not None:
                queue.put(1)

        if queue is not None:
            queue.put(None)

        return res

    @d.dedent
    def simulate_many(
        self,
        n_sims: int,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
        n_jobs: Optional[int] = None,
        backend: str = "loky",
        show_progress_bar: bool = True,
    ) -> List[np.ndarray]:
        """
        Simulate many random walks.

        Parameters
        ----------
        n_sims
            Number of random walks to simulate.
        %(rw_sim.params)s
        %(parallel)s

        Returns
        -------
        List of arrays of shape ``(max_iter + 1,)`` of states that have been visited. If ``stop_ixs`` was specified,
        the arrays may have smaller shape.
        """
        if n_sims <= 0:
            raise ValueError(
                f"Expected number of simulations to be positive, found `{n_sims}`."
            )
        max_iter = self._max_iter(max_iter)
        start = logg.info(
            f"Simulating `{n_sims}` random walks of maximum length `{max_iter}`"
        )

        simss = parallelize(
            self._simulate_many,
            collection=np.arange(n_sims),
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
            as_array=False,
            unit="sim",
        )(max_iter=max_iter, seed=seed, successive_hits=successive_hits)
        simss = list(chain.from_iterable(simss))

        logg.info("    Finish", time=start)

        return simss



import pytest

from cellrank.tl.kernels._random_walk import RandomWalk

import numpy as np


class TestRandomWalk:
    def test_matrix_not_row_stochastic(self, test_matrix_4: np.ndarray):
        with pytest.raises(
            ValueError, match=r"Transition matrix is not row-stochastic."
        ):
            _ = RandomWalk(test_matrix_4)

    def test_starting_dist_does_not_sum_to_1(self, test_matrix_1: np.ndarray):
        with pytest.raises(
            ValueError, match=r"No starting indices have been selected."
        ):
            _ = RandomWalk(test_matrix_1, start_ixs=[1000])

    @pytest.mark.parametrize("kind", ["simulations", "iterations", "hits"])
    def test_invalid_numbers(self, test_matrix_1: np.ndarray, kind: str):
        rw = RandomWalk(test_matrix_1)
        with pytest.raises(ValueError, match=kind):
            if kind == "simulations":
                rw.simulate_many(n_sims=0)
            elif kind == "iterations":
                rw.simulate_one(max_iter=1)
            elif kind == "hits":
                rw.simulate_one(successive_hits=-1)
            else:
                raise NotImplementedError(kind)

    @pytest.mark.parametrize("n_sims", [1, 10])
    def test_reproducibility(self, test_matrix_1: np.ndarray, n_sims: int):
        rw1 = RandomWalk(test_matrix_1)
        rw2 = RandomWalk(test_matrix_1)

        if n_sims == 1:
            r1, r2 = rw1.simulate_one(seed=42), rw2.simulate_one(seed=42)
        else:
            r1, r2 = rw1.simulate_many(10, seed=42), rw2.simulate_many(10, seed=42)

        np.testing.assert_array_equal(r1, r2)

    def test_simulate_one(self, test_matrix_1: np.ndarray):
        res = RandomWalk(test_matrix_1).simulate_one(max_iter=10)

        assert isinstance(res, np.ndarray)
        assert np.issubdtype(res.dtype, np.integer)
        np.testing.assert_array_equal(res.shape, (11,))

    def test_start_ixs(self, test_matrix_1: np.ndarray):
        res = RandomWalk(test_matrix_1, start_ixs=[1]).simulate_one(max_iter=10)

        assert res[0] == 1

    def test_stop_ixs(self, test_matrix_1: np.ndarray):
        res = RandomWalk(test_matrix_1, stop_ixs=[1]).simulate_one(
            max_iter=1000, seed=42
        )

        assert len(res) <= 1000
        assert res[-1] == 1
        assert res[-2] != 1

    def successive_hits(self, test_matrix_1: np.ndarray):
        res = RandomWalk(test_matrix_1, stop_ixs=[1]).simulate_one(
            max_iter=1000, seed=42, successive_hits=2
        )

        assert len(res) <= 1000
        assert res[-1] == 1
        assert res[-2] == 1
        assert res[-3] == 1
        assert res[-4] != 1

    @pytest.mark.parametrize("n_jobs", [1, 2])
    @pytest.mark.parametrize("backend", ["threading", "loky"])
    def test_simulate_many(self, test_matrix_1: np.ndarray, backend: str, n_jobs: int):
        res = RandomWalk(test_matrix_1).simulate_many(
            10, max_iter=100, seed=42, backend=backend, n_jobs=n_jobs
        )

        assert isinstance(res, list)
        assert len(res) == 10
        for i, r in enumerate(res):
            assert isinstance(r, np.ndarray), i
            assert np.issubdtype(r.dtype, np.integer), i
            np.testing.assert_array_equal(r.shape, (101,))



"""Graph pl module."""

from copy import deepcopy
from types import MappingProxyType
from typing import Dict, Tuple, Union, TypeVar, Callable, Optional, Sequence
from pathlib import Path

from cellrank import logging as logg
from cellrank.ul._docs import d
from cellrank.tl._utils import save_fig
from cellrank.ul._utils import _read_graph_data
from cellrank.tl._constants import _colors

import numpy as np
import pandas as pd
from scipy.sparse import issparse, spmatrix
from pandas.api.types import is_categorical_dtype

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap
from matplotlib.patches import ArrowStyle, FancyArrowPatch
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1 import make_axes_locatable

AnnData = TypeVar("AnnData")


KEYLOCS = str
KEYS = str
_msg_shown = False


@d.dedent
def graph(
    data: Union[AnnData, np.ndarray, spmatrix],
    graph_key: Optional[str] = None,
    ixs: Optional[Union[range, np.array]] = None,
    layout: Union[str, Dict, Callable] = "umap",
    keys: Sequence[KEYS] = ("incoming",),
    keylocs: Union[KEYLOCS, Sequence[KEYLOCS]] = "uns",
    node_size: float = 400,
    labels: Optional[Union[Sequence[str], Sequence[Sequence[str]]]] = None,
    top_n_edges: Optional[Union[int, Tuple[int, bool, str]]] = None,
    self_loops: bool = True,
    self_loop_radius_frac: Optional[float] = None,
    filter_edges: Optional[Tuple[float, float]] = None,
    edge_reductions: Union[Callable, Sequence[Callable]] = np.sum,
    edge_reductions_restrict_to_ixs: Optional[Union[range, np.ndarray]] = None,
    edge_weight_scale: float = 10,
    edge_width_limit: Optional[float] = None,
    edge_alpha: float = 1.0,
    edge_normalize: bool = False,
    edge_use_curved: bool = True,
    arrows: bool = True,
    font_size: int = 12,
    font_color: str = "black",
    color_nodes: bool = True,
    cat_cmap: ListedColormap = cm.Set3,
    cont_cmap: ListedColormap = cm.viridis,
    legend_loc: Optional[str] = "best",
    title: Optional[Union[str, Sequence[Optional[str]]]] = None,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, Path]] = None,
    layout_kwargs: Dict = MappingProxyType({}),
) -> None:
    """
    Plot a graph, visualizing incoming and outgoing edges or self-transitions.

    This is a utility function to look in more detail at the transition matrix in areas of interest, e.g. around an
    endpoint of development. This function is meant to visualise a small subset of nodes (~100-500) and the most likely
    transitions between them. Note that limiting edges visualized using ``top_n_edges`` will speed things up,
    as well as reduce the visual clutter.

    Parameters
    ----------
    data
        The graph data to be plotted.
    graph_key
        Key in ``adata.obsp`` or ``adata.uns`` where the graph is stored. Only used
        when ``data`` is :class:`~anndata.Anndata` object.
    ixs
        Subset of indices of the graph to visualize.
    layout
        Layout to use for graph drawing.

        - If :class:`str`, search for embedding in ``adata.obsm['X_{layout}']``.
          Use ``layout_kwargs={'components': [0, 1]}`` to select components.
        - If :class:`dict`, keys should be values in interval ``[0, len(ixs))``
          and values `(x, y)` pairs corresponding to node positions.
    keys
        Keys in ``adata.obs``, ``adata.obsm`` or ``adata.obsp`` to color the nodes.

        - If `'incoming'`, `'outgoing'` or `'self_loops'`, visualize reduction (see ``edge_reductions``)
          for each node based on incoming or outgoing edges, respectively.
    keylocs
        Locations of ``keys``. Can be any attribute of ``data`` if it's :class:`anndata.AnnData` object.
    node_size
        Size of the nodes.
    labels
        Labels of the nodes.
    top_n_edges
        Either top N outgoing edges in descending order or a tuple
        ``(top_n_edges, in_ascending_order, {'incoming', 'outgoing'})``. If `None`, show all edges.
    self_loops
        Whether visualize self transitions and also to consider them in ``top_n_edges``.
    self_loop_radius_frac
        Fraction of a unit circle to visualize self transitions. If `None`, use ``node_size / 1000``.
    filter_edges
        Whether to remove all edges not in `[min, max]` interval.
    edge_reductions
        Aggregation function to use when coloring nodes by edge weights.
    edge_reductions_restrict_to_ixs
        Whether to use the full graph when calculating the ``edge_reductions`` or just use the nodes
        marked by the ``ixs`` and this parameter. If `None`, it's the same as ``ixs``.
    edge_weight_scale
        Number by which to scale the width of the edges. Useful when the weights are small.
    edge_width_limit
        Upper bound for the width of the edges. Useful when weights are unevenly distributed.
    edge_alpha
        Alpha channel value for edges and arrows.
    edge_normalize
        If `True`, normalize edges to `[0, 1]` interval prior to applying any scaling or truncation.
    edge_use_curved
        If `True`, use curved edges. This can improve visualization at a small performance cost.
    arrows
        Whether to show the arrows. Setting this to `False` may dramatically speed things up.
    font_size
        Font size for node labels.
    font_color
        Label color of the nodes.
    color_nodes
        Whether to color the nodes
    cat_cmap
        Categorical colormap used when ``keys`` contain categorical variables.
    cont_cmap
        Continuous colormap used when ``keys`` contain continuous variables.
    legend_loc
        Location of the legend.
    title
        Title of the figure(s), one for each ``key``.
    %(plotting)s
    layout_kwargs
        Additional keyword arguments for ``layout``.

    Returns
    -------
    %(just_plots)s
    """

    from anndata import AnnData as _AnnData

    import networkx as nx

    def plot_arrows(curves, G, pos, ax, edge_weight_scale):
        for line, (edge, val) in zip(curves, G.edges.items()):
            if edge[0] == edge[1]:
                continue

            mask = (~np.isnan(line)).all(axis=1)
            line = line[mask, :]
            if not len(line):  # can be all NaNs
                continue

            line = line.reshape((-1, 2))
            X, Y = line[:, 0], line[:, 1]

            node_start = pos[edge[0]]
            # reverse
            if np.where(np.isclose(node_start - line, [0, 0]).all(axis=1))[0][0]:
                X, Y = X[::-1], Y[::-1]

            mid = len(X) // 2
            posA, posB = zip(X[mid : mid + 2], Y[mid : mid + 2])

            arrow = FancyArrowPatch(
                posA=posA,
                posB=posB,
                # we clip because too small values
                # cause it to crash
                arrowstyle=ArrowStyle.CurveFilledB(
                    head_length=np.clip(
                        val["weight"] * edge_weight_scale * 4,
                        _min_edge_weight,
                        edge_width_limit,
                    ),
                    head_width=np.clip(
                        val["weight"] * edge_weight_scale * 2,
                        _min_edge_weight,
                        edge_width_limit,
                    ),
                ),
                color="k",
                zorder=float("inf"),
                alpha=edge_alpha,
                linewidth=0,
            )
            ax.add_artist(arrow)

    def normalize_weights():
        weights = np.array([v["weight"] for v in G.edges.values()])
        minn = np.min(weights)
        weights = (weights - minn) / (np.max(weights) - minn)
        for v, w in zip(G.edges.values(), weights):
            v["weight"] = w

    def remove_top_n_edges():
        if top_n_edges is None:
            return

        if isinstance(top_n_edges, (tuple, list)):
            to_keep, ascending, group_by = top_n_edges
        else:
            to_keep, ascending, group_by = top_n_edges, False, "out"

        if group_by not in ("incoming", "outgoing"):
            raise ValueError(
                "Argument `groupby` in `top_n_edges` must be either `'incoming`' or `'outgoing'`."
            )

        source, target = zip(*G.edges)
        weights = [v["weight"] for v in G.edges.values()]
        tmp = pd.DataFrame({"outgoing": source, "incoming": target, "w": weights})

        if not self_loops:
            # remove self loops
            tmp = tmp[tmp["incoming"] != tmp["outgoing"]]

        to_keep = set(
            map(
                tuple,
                tmp.groupby(group_by)
                .apply(
                    lambda g: g.sort_values("w", ascending=ascending).take(
                        range(min(to_keep, len(g)))
                    )
                )[["outgoing", "incoming"]]
                .values,
            )
        )

        for e in list(G.edges):
            if e not in to_keep:
                G.remove_edge(*e)

    def remove_low_weight_edges():
        if filter_edges is None or filter_edges == (None, None):
            return

        minn, maxx = filter_edges
        minn = minn if minn is not None else -np.inf
        maxx = maxx if maxx is not None else np.inf

        for e, attr in list(G.edges.items()):
            if attr["weight"] < minn or attr["weight"] > maxx:
                G.remove_edge(*e)

    _min_edge_weight = 0.00001

    if edge_width_limit is None:
        logg.debug("Not limiting width of edges")
        edge_width_limit = float("inf")

    if self_loop_radius_frac is None:
        self_loop_radius_frac = (
            node_size / 2000 if node_size >= 200 else node_size / 1000
        )
        logg.debug(f"Setting self loop radius fraction to `{self_loop_radius_frac}`")

    if not isinstance(keys, (tuple, list)):
        keys = [keys]

    if not isinstance(keylocs, (tuple, list)):
        keylocs = [keylocs] * len(keys)
    elif len(keylocs) == 1:
        keylocs = keylocs * 3
    elif all(map(lambda k: k in ("incoming", "outgoing", "self_loops"), keys)):
        # don't care about keylocs since they are irrelevant
        logg.debug("Ignoring key locations")
        keylocs = [None] * len(keys)

    if not isinstance(edge_reductions, (tuple, list)):
        edge_reductions = [edge_reductions] * len(keys)
    if not all(map(callable, edge_reductions)):
        raise ValueError("Not all `edge_reductions` functions are callable.")

    if not isinstance(labels, (tuple, list)):
        labels = [labels] * len(keys)
    elif not len(labels):
        labels = [None] * len(keys)
    elif not isinstance(labels[0], (tuple, list)):
        labels = [labels] * len(keys)

    if len(keys) != len(labels):
        raise ValueError(
            f"`Keys` and `labels` must be of the same shape, found `{len(keys)}` and `{len(labels)}`."
        )
    if title is None or isinstance(title, str):
        title = [title] * len(keys)
    if len(title) != len(keys):
        raise ValueError(
            f"`Titles` and `keys` must be of the same shape, found `{len(title)}` and `{len(keys)}`."
        )

    if isinstance(data, _AnnData):
        if graph_key is None:
            raise ValueError(
                "Argument `graph_key` cannot be `None` when `data` is `anndata.Anndata` object."
            )
        gdata = _read_graph_data(data, graph_key)
    elif isinstance(data, (np.ndarray, spmatrix)):
        gdata = data
    else:
        raise TypeError(
            f"Expected argument `data` to be one of `anndata.AnnData`, `numpy.ndarray`, `scipy.sparse.spmatrix`, "
            f"found `{type(data).__name__!r}`."
        )
    is_sparse = issparse(gdata)

    gdata_full = gdata
    if ixs is not None:
        gdata = gdata[ixs, :][:, ixs]
    else:
        ixs = list(range(gdata.shape[0]))

    if edge_reductions_restrict_to_ixs is None:
        edge_reductions_restrict_to_ixs = ixs

    start = logg.info("Creating graph")
    G = (
        nx.from_scipy_sparse_matrix(gdata, create_using=nx.DiGraph)
        if is_sparse
        else nx.from_numpy_array(gdata, create_using=nx.DiGraph)
    )

    remove_low_weight_edges()
    remove_top_n_edges()
    if edge_normalize:
        normalize_weights()
    logg.info("    Finish", time=start)

    # do NOT recreate the graph, for the edge reductions
    # gdata = nx.to_numpy_array(G)

    if figsize is None:
        figsize = (12, 8 * len(keys))

    fig, axes = plt.subplots(nrows=len(keys), ncols=1, figsize=figsize, dpi=dpi)
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    axes = np.ravel(axes)

    if isinstance(layout, str):
        if f"X_{layout}" not in data.obsm:
            raise KeyError(f"Unable to find embedding `'X_{layout}'` in `adata.obsm`.")
        components = layout_kwargs.get("components", [0, 1])
        if len(components) != 2:
            raise ValueError(
                f"Components in `layout_kwargs` must be of length `2`, found `{len(components)}`."
            )
        emb = data.obsm[f"X_{layout}"][:, components]
        pos = {i: emb[ix, :] for i, ix in enumerate(ixs)}
        logg.info(f"Embedding graph using `{layout!r}` layout")
    elif isinstance(layout, dict):
        rng = range(len(ixs))
        for k, v in layout.items():
            if k not in rng:
                raise ValueError(
                    f"Key in `layout` must be in `range(len(ixs))`, found `{k}`."
                )
            if len(v) != 2:
                raise ValueError(
                    f"Value in `layout` must be a `tuple` or a `list` of length 2, found `{len(v)}`."
                )
        pos = layout
        logg.debug("Using precomputed layout")
    elif callable(layout):
        start = logg.info(f"Embedding graph using `{layout.__name__!r}` layout")
        pos = layout(G, **layout_kwargs)
        logg.info("    Finish", time=start)
    else:
        raise TypeError(
            f"Argument `layout` must be either a `string`, "
            f"a `dict` or a `callable`, found `{type(layout)}`."
        )

    curves, lc = None, None
    if edge_use_curved:
        try:
            from cellrank.pl._utils import _curved_edges

            logg.debug("Creating curved edges")
            curves = _curved_edges(G, pos, self_loop_radius_frac, polarity="directed")
            lc = LineCollection(
                curves,
                colors="black",
                linewidths=np.clip(
                    np.ravel([v["weight"] for v in G.edges.values()])
                    * edge_weight_scale,
                    0,
                    edge_width_limit,
                ),
                alpha=edge_alpha,
            )
        except ImportError as e:
            global _msg_shown
            if not _msg_shown:
                print(
                    str(e)[:-1],
                    "in order to use curved edges or specify `edge_use_curved=False`.",
                )
                _msg_shown = True

    for ax, keyloc, title, key, labs, er in zip(
        axes, keylocs, title, keys, labels, edge_reductions
    ):
        label_col = {}  # dummy value

        if key in ("incoming", "outgoing", "self_loops"):
            if key in ("incoming", "outgoing"):
                axis = int(key == "outgoing")
                tmp_data = (
                    gdata_full[ixs, :][:, edge_reductions_restrict_to_ixs]
                    if axis
                    else gdata_full[edge_reductions_restrict_to_ixs, :][:, ixs]
                )
                vals = er(tmp_data, axis=axis)
            else:
                vals = gdata.diagonal() if is_sparse else np.diag(gdata)

            if issparse(vals):
                vals = vals.toarray()
            vals = np.squeeze(np.array(vals))

            assert len(pos) == len(vals), (
                f"Expected reduction values to be of length `{len(pos)}`, "
                f"found `{len(vals)}`."
            )

            node_v = dict(zip(pos.keys(), vals))
        else:
            label_col = getattr(data, keyloc)
            if key in label_col:
                node_v = dict(zip(pos.keys(), label_col[key]))
            else:
                raise RuntimeError(f"Key `{key!r}` not found in `adata.{keyloc}`.")

        if labs is not None:
            if len(labs) != len(pos):
                raise RuntimeError(
                    f"Number of labels ({len(labels)}) and nodes ({len(pos)}) mismatch."
                )
            nx.draw_networkx_labels(
                G,
                pos,
                labels=labs if isinstance(labs, dict) else dict(zip(pos.keys(), labs)),
                ax=ax,
                font_color=font_color,
                font_size=font_size,
            )

        if lc is not None and curves is not None:
            ax.add_collection(deepcopy(lc))  # copying necessary
            if arrows:
                plot_arrows(curves, G, pos, ax, edge_weight_scale)
        else:
            nx.draw_networkx_edges(
                G,
                pos,
                width=[
                    np.clip(
                        v["weight"] * edge_weight_scale,
                        _min_edge_weight,
                        edge_width_limit,
                    )
                    for _, v in G.edges.items()
                ],
                alpha=edge_alpha,
                edge_color="black",
                arrows=True,
                arrowstyle="-|>",
            )

        if key in label_col and is_categorical_dtype(label_col[key]):
            values = label_col[key]
            if keyloc in ("obs", "obsm"):
                values = values[ixs]
            categories = values.cat.categories
            color_key = _colors(key)
            if color_key in data.uns:
                mapper = dict(zip(categories, data.uns[color_key]))
            else:
                mapper = dict(
                    zip(categories, map(cat_cmap.get, range(len(categories))))
                )

            colors = []
            seen = set()

            for v in values:
                colors.append(mapper[v])
                seen.add(v)

            nodes_kwargs = dict(cmap=cat_cmap, node_color=colors)  # noqa
            if legend_loc not in ("none", None):
                x, y = pos[0]
                for label in sorted(seen):
                    ax.plot([x], [y], label=label, color=mapper[label])
                ax.legend(loc=legend_loc)
        else:
            values = list(node_v.values())
            vmin, vmax = np.min(values), np.max(values)
            nodes_kwargs = dict(  # noqa
                cmap=cont_cmap, node_color=values, vmin=vmin, vmax=vmax
            )

            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="1.5%", pad=0.05)
            _ = mpl.colorbar.ColorbarBase(
                cax,
                cmap=cont_cmap,
                norm=mpl.colors.Normalize(vmin=vmin, vmax=vmax),
                ticks=np.linspace(vmin, vmax, 10),
            )

        if color_nodes is False:
            nodes_kwargs = {}

        nx.draw_networkx_nodes(G, pos, node_size=node_size, ax=ax, **nodes_kwargs)

        ax.set_title(key if title is None else title)
        ax.axis("off")

    if save is not None:
        save_fig(fig, save)

    fig.show()



"""Base properties used within the estimators."""
import sys
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Dict, List, Tuple, Union, Iterable, Optional
from inspect import isabstract

import scvelo as scv
import cellrank.logging as logg
from anndata import AnnData
from cellrank.ul._docs import d, _initial, _terminal
from cellrank.tl._utils import RandomKeys, _make_cat, _partition, _complex_warning
from cellrank.ul._utils import Pickleable
from cellrank.tl._colors import _create_categorical_colors
from cellrank.tl.kernels import PrecomputedKernel
from cellrank.tl._lineage import Lineage
from cellrank.tl._constants import Direction, DirPrefix, DirectionPlot
from cellrank.tl.kernels._utils import _filter_kwargs
from cellrank.tl.estimators._utils import (
    Metadata,
    Sequence,
    _create_property,
    singledispatchmethod,
    _delegate_method_dispatch,
)
from cellrank.tl.kernels._base_kernel import KernelExpression
from cellrank.tl.estimators._constants import META_KEY, A, F, P

import numpy as np
import pandas as pd
from scipy.sparse import issparse, spmatrix
from pandas.api.types import is_categorical_dtype

import matplotlib as mpl
from matplotlib import cm


def is_abstract(classname: str) -> bool:  # TODO: determine the necessity of this
    """
    Check whether class with a given name inside this module is abstract.

    Parameters
    ----------
    classname
        Name of the class.

    Returns
    -------
    bool
        `True` if the class is abstract, otherwise `False`.

    """

    cls = getattr(sys.modules[__name__], classname, None)
    return cls is not None and isabstract(cls)


class PropertyMeta(ABCMeta, type):
    """Metaclass for all the properties."""

    @staticmethod
    def update_attributes(md: Metadata, attributedict: Dict[str, Any]) -> Optional[str]:
        """
        Update ``attributedict`` with new attribute and property.

        Parameters
        ----------
        md
            Metadata object, containing attribute name, property names, etc.
        attributedict
            Dictionary with attributes.

        Returns
        -------
        str or None
            Old property name, if found, otherwise a newly constructed one, or `None` if no property is desired.
        """

        # TODO: determine whether supporting strings is such a good idea
        if not isinstance(md, Metadata):
            raise TypeError(
                f"Expected a `Metadata` object, found `{type(md).__name__}`."
            )

        if not isinstance(md.attr, (str, A)):
            raise TypeError(
                f"Attribute `{md.attr}` must be of type `A` or `str`, found `{type(md.attr).__name__!r}`."
            )
        if not str(md.attr).startswith("_"):
            raise ValueError(f"Attribute `{md.attr!r}` must start with `'_'`.")

        attributedict[str(md.attr)] = md.default

        if md.prop == P.NO_PROPERTY:
            return

        if not isinstance(md.prop, (str, P)):
            raise TypeError(
                f"Property must be of type `P` or `str`, found `{type(md.attr).__name__!r}`."
            )

        prop_name = str(md.attr).lstrip("_") if md.prop == P.EMPTY else str(md.prop)

        if not len(prop_name):
            raise ValueError(f"Property name for attribute `{md.attr}` is empty.")
        if prop_name.startswith("_"):
            raise ValueError(
                f"Property musn't start with an underscore: `{prop_name!r}`."
            )

        attributedict[prop_name] = _create_property(
            str(md.attr), prop_name, doc=md.doc, return_type=md.dtype
        )

        return prop_name

    def __new__(cls, clsname, superclasses, attributedict):
        """
        Create a new instance.

        Parameters
        ----------
        clsname
            Name of class to be constructed.
        superclasses
            List of superclasses.
        attributedict
            Dictionary of attributes.
        """

        compute_md, metadata = attributedict.pop(META_KEY, None), []

        if compute_md is None:
            return super().__new__(cls, clsname, superclasses, attributedict)

        if isinstance(compute_md, str):
            compute_md = Metadata(attr=compute_md)
        elif not isinstance(compute_md, (tuple, list)):
            raise TypeError(
                f"Expected property metadata to be `list` or `tuple`,"
                f"found `{type(compute_md).__name__!r}`."
            )
        elif len(compute_md) == 0:
            raise ValueError("No metadata found.")
        else:
            compute_md, *metadata = (
                Metadata(attr=md) if isinstance(md, str) else md for md in compute_md
            )

        prop_name = PropertyMeta.update_attributes(compute_md, attributedict)
        plot_name = str(compute_md.plot_fmt).format(prop_name)

        if compute_md.compute_fmt != F.NO_FUNC:
            if "_compute" in attributedict:
                attributedict[
                    str(compute_md.compute_fmt).format(prop_name)
                ] = attributedict["_compute"]

        if (
            compute_md.plot_fmt != F.NO_FUNC
            and VectorPlottable in superclasses
            and plot_name not in attributedict
            and not is_abstract(clsname)
        ):
            raise TypeError(
                f"Method `{plot_name}` is not implemented for class `{clsname}`."
            )

        for md in metadata:
            PropertyMeta.update_attributes(md, attributedict)

        res = super().__new__(cls, clsname, superclasses, attributedict)

        if compute_md.plot_fmt != F.NO_FUNC and Plottable in res.mro():
            # _this is intended singledispatchmethod
            # unfortunately, `_plot` is not always in attributedict, so we can't just check for it
            # and res._plot is just a regular function
            # if this gets buggy in the future, consider switching from singlemethoddispatch
            setattr(
                res,
                plot_name,
                _delegate_method_dispatch(res._plot, "_plot", prop_name, skip=2),
            )

        return res


class Property(ABC, metaclass=PropertyMeta):
    """Base class for all the properties."""


class KernelHolder(ABC):
    """Base class which holds a :class:`cellrank.tool.kernels._kernel.KernelExpression`."""

    def __init__(
        self,
        obj: Union[AnnData, np.ndarray, spmatrix, KernelExpression],
        key: Optional[str] = None,
        obsp_key: Optional[str] = None,
        write_to_adata: bool = True,
    ):
        if isinstance(obj, KernelExpression):
            self._kernel = obj
        elif isinstance(obj, (np.ndarray, spmatrix)):
            self._kernel = PrecomputedKernel(obj)
        elif isinstance(obj, AnnData):
            if obsp_key is None:
                raise ValueError(
                    "Specify `obsp_key=...` when supplying an `AnnData` object."
                )
            elif obsp_key not in obj.obsp.keys():
                raise KeyError(f"Key `{obsp_key!r}` not found in `adata.obsp`.")
            self._kernel = PrecomputedKernel(obj.obsp[obsp_key], adata=obj)
        else:
            raise TypeError(
                f"Expected an object of type `KernelExpression`, `numpy.ndarray`, `scipy.sparse.spmatrix` "
                f"or `anndata.AnnData`, got `{type(obj).__name__!r}`."
            )

        if self.kernel._transition_matrix is None:
            # access the private attribute to avoid accidentally computing the transition matrix
            # in principle, it doesn't make a difference, apart from not seeing the message
            logg.warning("Computing transition matrix using the default parameters")
            self.kernel.compute_transition_matrix()

        if write_to_adata:
            self.kernel.write_to_adata(key=key)

    @property
    def _direction(self):
        return Direction.BACKWARD if self.kernel.backward else Direction.FORWARD

    @property
    def transition_matrix(self) -> Union[np.ndarray, spmatrix]:
        """Transition matrix."""
        return self.kernel.transition_matrix

    @property
    def issparse(self) -> bool:
        """Whether the transition matrix is sparse or not."""
        return issparse(self.transition_matrix)

    @property
    def kernel(self) -> KernelExpression:
        """Underlying kernel."""
        return self._kernel

    @property
    @d.dedent
    def adata(self) -> AnnData:
        """
        Annotated data object.

        Returns
        -------
        %(adata_ret)s
        """
        return self.kernel.adata

    def __len__(self):
        return self.kernel.transition_matrix.shape[0]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}[n={len(self)}, kernel={repr(self.kernel)}]"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}[n={len(self)}, kernel={str(self.kernel)}]"


class VectorPlottable(KernelHolder, Property):
    """
    Injector class which plots vectors.

    To be used in conjunction with:

        - :class:`cellrank.tool.estimators._decomposition.Eig`
        - :class:`cellrank.tool.estimators._decomposition.Schur`
    """

    @d.get_sections(base="plot_vectors", sections=["Parameters", "Returns"])
    @d.dedent
    def _plot_vectors(
        self,
        vectors: Optional[np.ndarray],
        prop: str,
        use: Optional[Union[int, Tuple[int], List[int]]] = None,
        abs_value: bool = False,
        cluster_key: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Plot vectors in an embedding.

        Parameters
        ----------
        use
            Which or how many vectors are to be plotted.
        abs_value
            Whether to take the absolute value before plotting.
        cluster_key
            Key in :attr:`adata` ``.obs`` for plotting categorical observations.
        %(basis)s
        kwargs
            Keyword arguments for :func:`scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """

        if prop not in (P.EIG.v, P.SCHUR.v):
            raise ValueError(
                f"Invalid kind `{prop!r}`. Valid options are `{P.EIG!r}`, `{P.SCHUR!r}``."
            )
        if vectors is None:
            raise RuntimeError(f"Compute `.{prop}` first as `{F.COMPUTE.fmt(prop)}()`.")

        if prop == P.SCHUR.s:
            is_schur = True
            name = "Schur "
        else:
            is_schur = False
            name = "eigen"

        # check whether dimensions are consistent
        if self.adata.n_obs != vectors.shape[0]:
            raise ValueError(
                f"Number of cells ({self.adata.n_obs}) is inconsistent with the first"
                f"dimension of vectors ({vectors.shape[0]})."
            )

        if use is None:
            m = (
                getattr(self, P.EIG.s).get("eigengap", vectors.shape[1]) + 1
                if hasattr(self, P.EIG.s) and not is_schur
                else vectors.shape[1]
            )
            use = list(range(is_schur, m))
        elif isinstance(use, int):
            use = list(range(is_schur, use + is_schur))
        elif not isinstance(use, (tuple, list, range)):
            raise TypeError(
                f"Argument `use` must be either `int`, `tuple`, `list` or `range`,"
                f"found `{type(use).__name__}`."
            )
        else:
            if not all(map(lambda u: isinstance(u, int), use)):
                raise TypeError("Not all values in `use` argument are integers.")
        use = list(use)
        if not use:
            raise ValueError("Nothing to plot.")

        muse = max(use)
        if muse >= vectors.shape[1]:
            raise ValueError(
                f"Maximum specified {name}vector ({muse}) is larger "
                f"than the number of computed {name}vectors ({vectors.shape[1]})."
            )
        V_ = vectors[:, use]

        if is_schur:
            title = [f"{name}vector {i}" for i in use]
        else:
            D = kwargs.pop("D")
            V_ = _complex_warning(V_, use, use_imag=kwargs.pop("use_imag", False))
            title = [fr"$\lambda_{i}$={d:.02f}" for i, d in zip(use, D[use])]

        if abs_value:
            V_ = np.abs(V_)

        color = list(V_.T)
        if cluster_key is not None:
            color = [cluster_key] + color
        cmap = kwargs.pop("cmap", "viridis")

        logg.debug(f"Plotting `{use}` {name}vectors")

        scv.pl.scatter(self.adata, color=color, title=title, cmap=cmap, **kwargs)


class Plottable(KernelHolder, Property):
    """
    Injector which plots macrostates or terminal states or absorption probabilities.

    To be used in conjunction with:

        - :class:`cellrank.tool.estimators._property.MacroStates`.
        - :class:`cellrank.tool.estimators._property.TermStates`.
        - :class:`cellrank.tool.estimators._property.AbsProbs`.
    """

    @d.get_sections(base="plot_discrete")
    @d.dedent
    def _plot_discrete(
        self,
        data: pd.Series,
        prop: str,
        lineages: Optional[Union[str, Sequence[str]]] = None,
        cluster_key: Optional[str] = None,
        same_plot: bool = True,
        title: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ) -> None:
        """
        Plot the states for each uncovered lineage.

        Parameters
        ----------
        lineages
            Plot only these lineages. If `None`, plot all lineages.
        cluster_key
            Key from :attr:`adata` ``.obs`` for plotting categorical observations.
        same_plot
            Whether to plot the lineages on the same plot or separately.
        title
            The title of the plot.
        %(basis)s
        kwargs
            Keyword arguments for :func:`scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """

        if data is None:
            raise RuntimeError(
                f"Compute `.{prop}` first as `.{F.COMPUTE.fmt(prop)}()`."
            )
        if not is_categorical_dtype(data):
            raise TypeError(
                f"Expected property `.{prop}` to be categorical, found `{type(data).__name__!r}`."
            )
        if prop in (P.ABS_PROBS.s, P.TERM.s):
            colors = getattr(self, A.TERM_COLORS.v, None)
        elif prop == P.MACRO.v:
            colors = getattr(self, A.MACRO_COLORS.v, None)
        else:
            logg.debug("No colors found. Creating new ones")
            colors = _create_categorical_colors(len(data.cat.categories))
        colors = dict(zip(data.cat.categories, colors))

        if (
            lineages is not None
        ):  # these are states per-se, but I want to keep the arg names for dispatch the same
            if isinstance(lineages, str):
                lineages = [lineages]
            for state in lineages:
                if state not in data.cat.categories:
                    raise ValueError(
                        f"Invalid state `{state!r}`. Valid options are `{list(data.cat.categories)}`."
                    )
            data = data.copy()
            to_remove = list(set(data.cat.categories) - set(lineages))

            if len(to_remove) == len(data.cat.categories):
                raise RuntimeError(
                    "Nothing to plot because empty subset has been selected."
                )

            for state in to_remove:
                data[data == state] = np.nan
            data.cat.remove_categories(to_remove, inplace=True)

        if cluster_key is None:
            cluster_key = []
        elif isinstance(cluster_key, str):
            cluster_key = [cluster_key]
        if not isinstance(cluster_key, list):
            cluster_key = list(cluster_key)

        same_plot = same_plot or len(data.cat.categories) == 1
        kwargs["legend_loc"] = kwargs.get("legend_loc", "on data")

        with RandomKeys(
            self.adata, None if same_plot else len(data.cat.categories), where="obs"
        ) as keys:
            if same_plot:
                key = keys[0]
                self.adata.obs[key] = data
                self.adata.uns[f"{key}_colors"] = [
                    colors[c] for c in data.cat.categories
                ]

                if title is None:
                    title = (
                        f"{prop.replace('_', ' ')} "
                        f"({Direction.BACKWARD if self.kernel.backward else Direction.FORWARD})"
                    )
                if isinstance(title, str):
                    title = [title]

                scv.pl.scatter(
                    self.adata,
                    title=cluster_key + title,
                    color=cluster_key + keys,
                    **_filter_kwargs(scv.pl.scatter, **kwargs),
                )
            else:
                for key, cat in zip(keys, data.cat.categories):
                    d = data.copy()
                    d[data != cat] = None
                    d.cat.set_categories([cat], inplace=True)

                    self.adata.obs[key] = d
                    self.adata.uns[f"{key}_colors"] = [colors[cat]]

                scv.pl.scatter(
                    self.adata,
                    color=cluster_key + keys,
                    title=(
                        cluster_key
                        + [
                            f"{_initial if self.kernel.backward else _terminal} state {c}"
                            for c in data.cat.categories
                        ]
                    )
                    if title is None
                    else title,
                    **_filter_kwargs(scv.pl.scatter, **kwargs),
                )

    @d.get_sections(base="plot_continuous")
    @d.dedent
    def _plot_continuous(
        self,
        probs: Optional[Lineage],
        prop: str,
        lineages: Optional[Union[str, Iterable[str]]] = None,
        cluster_key: Optional[str] = None,
        mode: str = "embedding",
        time_key: str = "latent_time",
        title: Optional[str] = None,
        same_plot: bool = True,
        cmap: Union[str, mpl.colors.ListedColormap] = cm.viridis,
        **kwargs,
    ) -> None:
        """
        Plot continuous observations such as macrostates memberships or lineages in an embedding.

        Parameters
        ----------
        lineages
            Plot only these lineages. If `None`, plot all lineages.
        cluster_key
            Key from :attr:`adata` ``.obs`` for plotting categorical observations.
        %(time_mode)s
        time_key
            Key from :attr:`adata` ``.obs`` to use as a pseudotime ordering of the cells.
        title
            Either `None`, in which case titles are ``'{to,from} {terminal,initial} {state}'``,
            or an array of titles, one per lineage.
        same_plot
            Whether to plot the lineages on the same plot using color gradients when ``mode='embedding'``.
        cmap
            Colormap to use.
        %(basis)s
        kwargs
            Keyword arguments for :func:`scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """

        if probs is None:
            raise RuntimeError(
                f"Compute `.{prop}` first as `.{F.COMPUTE.fmt(prop)}()`."
            )

        if isinstance(lineages, str):
            lineages = [lineages]

        if lineages is None:
            lineages = probs.names
            A = probs
        else:
            A = probs[lineages]

        if not len(lineages):
            raise RuntimeError(
                "Nothing to plot because empty subset has been selected."
            )

        prefix = DirPrefix.BACKWARD if self.kernel.backward else DirPrefix.FORWARD
        same_plot = same_plot and mode == "embedding"  # set this silently

        A = A.copy()  # the below code modifies stuff inplace
        X = A.X  # list(A.T) behaves differently, because it's Lineage

        if X.shape[1] == 1:
            same_plot = (
                False  # because color_gradients for 1 state is buggy (looks empty)
            )
            # this is the case for only 1 recurrent class - all cells have prob. 1 of going there
            # however, matplotlib's plotting really picks up the slightest differences in the colormap, here we set
            # everything to one, if applicable
            if np.allclose(X, 1.0):
                X = np.ones_like(X)

        for col in X.T:
            mask = ~np.isclose(col, 1.0)
            # change the maximum value - the 1 is artificial and obscures the color scaling
            if np.sum(mask):
                max_not_one = np.max(col[mask])
                col[~mask] = max_not_one

        if mode == "time":
            if time_key not in self.adata.obs.keys():
                raise KeyError(f"Time key `{time_key!r}` not found in `adata.obs`.")
            time = self.adata.obs[time_key]
            if cluster_key is not None:
                logg.warning(
                    f"Cluster key `{cluster_key!r}` is ignored when `mode='time'`"
                )
            cluster_key = None

        color = list(X.T)
        if title is None:
            if same_plot:
                title = [
                    f"{prop.replace('_', ' ')} "
                    f"({DirectionPlot.BACKWARD if self.kernel.backward else Direction.FORWARD})"
                ]
            else:
                title = [f"{prefix} {lin}" for lin in lineages]
        elif isinstance(title, str):
            title = [title]

        if isinstance(cluster_key, str):
            cluster_key = [cluster_key]
        elif cluster_key is None:
            cluster_key = []
        if not isinstance(cluster_key, list):
            cluster_key = list(cluster_key)

        if not same_plot:
            color = cluster_key + color
            title = cluster_key + title

        if mode == "embedding":
            if same_plot:
                # to complement: https://github.com/theislab/scvelo/blob/master/scvelo/plotting/scatter.py#L269
                sorted_idx = np.argsort(X, axis=1)[:, ::-1][:, :2]
                pairs, cnts = np.unique(
                    np.sort(sorted_idx, axis=0), axis=0, return_counts=True
                )
                absent = set(A.names) - set(A.names[pairs[cnts > 1].flatten()])
                if absent:
                    # can't print which because it would be inaccurate, i.e. we would print the legend
                    # is missing when in fact, it would be visible
                    # this has been checked exactly according to scVelo's logic
                    logg.warning("Legend for some lineages may be missing")

                kwargs["color_gradients"] = A
                if len(cluster_key):
                    logg.warning(
                        "Ignoring `cluster_key` when plotting continuous observations in the same plot"
                    )
                # kwargs["color"] = cluster_key  this results in a bug, cluster_key data is overwritten, will make a PR
            else:
                kwargs["color"] = color

            if probs.shape[1] == 1 and prop in (P.MACRO_MEMBER.s, P.TERM_PROBS.s):
                if "perc" not in kwargs:
                    logg.warning(
                        "Did not detect percentile for stationary distribution. Setting `perc=[0, 95]`"
                    )
                    kwargs["perc"] = [0, 95]
                kwargs["color"] = X
                kwargs.pop("color_gradients", None)

            scv.pl.scatter(
                self.adata,
                title=title,
                color_map=cmap,
                **_filter_kwargs(scv.pl.scatter, **kwargs),
            )
        elif mode == "time":
            scv.pl.scatter(
                self.adata,
                x=time,
                color_map=cmap,
                y=color,
                title=title,
                xlabel=[time_key] * len(title),
                ylabel=["probability"] * len(title),
                **_filter_kwargs(scv.pl.scatter, **kwargs),
            )
        else:
            raise ValueError(
                f"Invalid mode `{mode!r}`. Valid options are: `'embedding'` or `'time'`."
            )

    @singledispatchmethod
    @d.dedent
    def _plot(
        self,
        data,
        prop: str,
        discrete: bool = False,
        lineages: Optional[Union[str, Sequence[str]]] = None,
        cluster_key: Optional[str] = None,
        mode: str = "embedding",
        time_key: str = "latent_time",
        title: Optional[str] = None,
        same_plot: bool = False,
        cmap: Union[str, mpl.colors.ListedColormap] = "viridis",
        **kwargs,
    ) -> None:
        """
        Plot discrete states or probabilities in an embedding.

        Parameters
        ----------
        discrete
            Whether to plot in discrete or continuous mode.
        %(plot_continuous.parameters)s

        Returns
        -------
        %(just_plots)s
        """
        # we keep all the arguments because from this function, we get the docs + signature
        raise RuntimeError(f"Compute `.{prop}` first as `.{F.COMPUTE.fmt(prop)}()`.")

    @_plot.register(pd.Series)
    def _(self, data: pd.Series, prop: str, discrete: bool = False, **kwargs) -> None:
        if discrete and kwargs.get("mode", "embedding") == "time":
            logg.warning(
                "`mode='time'` is implemented in continuous case, plotting in continuous mode"
            )
            discrete = False

        if discrete:
            self._plot_discrete(data, prop, **kwargs)
        elif prop == P.MACRO.v:  # GPCCA
            prop = P.MACRO_MEMBER.v
            self._plot_continuous(getattr(self, prop, None), prop, **kwargs)
        elif prop == P.TERM.v:
            probs = getattr(self, A.TERM_ABS_PROBS.s, None)
            # we have this only in GPCCA
            if isinstance(probs, Lineage):
                self._plot_continuous(probs, P.TERM_PROBS.v, **kwargs)
            else:
                logg.warning(
                    f"Unable to plot continuous observations for `{prop!r}`, plotting in discrete mode"
                )
                self._plot_discrete(data, prop, **kwargs)
        else:
            raise NotImplementedError(
                f"Unable to plot property `.{prop}` in discrete mode."
            )

    @_plot.register(Lineage)
    def _(self, data: Lineage, prop: str, discrete: bool = False, **kwargs) -> None:
        if discrete and kwargs.get("mode", "embedding") == "time":
            logg.warning(
                "`mode='time'` is implemented in continuous case, plotting in continuous mode"
            )
            discrete = False

        if not discrete:
            self._plot_continuous(data, prop, **kwargs)
        elif prop == P.ABS_PROBS.v:
            # for discrete and abs. probs, plot the terminal states
            prop = P.TERM.v
            self._plot_discrete(getattr(self, prop, None), prop, **kwargs)
        else:
            raise NotImplementedError(
                f"Unable to plot property `.{prop}` in continuous mode."
            )


class Macrostates(Plottable):
    """Class dealing with macrostates."""

    __prop_metadata__ = [
        Metadata(attr=A.MACRO, prop=P.MACRO, dtype=pd.Series),
        Metadata(
            attr=A.MACRO_MEMBER,
            prop=P.MACRO_MEMBER,
            dtype=Lineage,
        ),
        Metadata(attr=A.MACRO_COLORS, prop=P.NO_PROPERTY, dtype=np.ndarray),
    ]

    @abstractmethod
    def compute_macrostates(self, *args, **kwargs) -> None:  # noqa
        pass


class TerminalStates(Plottable):
    """Class dealing with terminal states."""

    __prop_metadata__ = [
        Metadata(attr=A.TERM, prop=P.TERM, dtype=pd.Series, doc="Terminal states."),
        Metadata(
            attr=A.TERM_PROBS,
            prop=P.TERM_PROBS,
            dtype=pd.Series,
            doc="Terminal states probabilities.",
        ),
        Metadata(attr=A.TERM_COLORS, prop=P.NO_PROPERTY, dtype=np.ndarray),
    ]

    @abstractmethod
    def set_terminal_states(self, *args: Any, **kwargs: Any) -> None:  # noqa
        pass

    @abstractmethod
    def compute_terminal_states(self, *args: Any, **kwargs: Any) -> None:  # noqa
        pass

    @abstractmethod
    def _write_terminal_states(self, *args: Any, **kwargs: Any) -> None:
        pass


class AbsProbs(Plottable):
    """Class dealing with absorption probabilities."""

    __prop_metadata__ = [
        Metadata(
            attr=A.ABS_PROBS,
            prop=P.ABS_PROBS,
            dtype=Lineage,
            doc="Absorption probabilities.",
        ),
        Metadata(
            attr=A.PRIME_DEG,
            prop=P.PRIME_DEG,
            dtype=pd.Series,
            doc="Priming degree.",
        ),
        Metadata(attr=A.LIN_ABS_TIMES, prop=P.LIN_ABS_TIMES, dtype=pd.DataFrame),
    ]

    @abstractmethod
    def _write_absorption_probabilities(self, *args, **kwargs) -> None:
        pass


class LinDrivers(Plottable):  # noqa
    __prop_metadata__ = [
        Metadata(
            attr=A.LIN_DRIVERS,
            prop=P.LIN_DRIVERS,
            dtype=pd.DataFrame,
            doc="Lineage drivers.",
            plot_fmt=F.NO_FUNC,
            # in essence ignore Plottable (could be done by registering DataFrame, but it's ugly
        )
    ]


class Partitioner(KernelHolder, ABC):
    """Abstract base class for partitioning transition matrix into sets of recurrent and transient states."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._is_irreducible = None  # no need to refactor these
        self._rec_classes = None
        self._trans_classes = None

    def compute_partition(self) -> None:
        """
        Compute communication classes for the Markov chain.

        Returns
        -------
        None
            Nothing, but updates the following fields:

                - :attr:`recurrent_classes`
                - :attr:`transient_classes`
                - :attr:`is_irreducible`
        """

        start = logg.info("Computing communication classes")
        n_states = len(self)

        rec_classes, trans_classes = _partition(self.transition_matrix)

        self._is_irreducible = len(rec_classes) == 1 and len(trans_classes) == 0

        if not self._is_irreducible:
            self._trans_classes = _make_cat(
                trans_classes, n_states, self.adata.obs_names
            )
            self._rec_classes = _make_cat(rec_classes, n_states, self.adata.obs_names)
            logg.info(
                f"Found `{(len(rec_classes))}` recurrent and `{len(trans_classes)}` transient classes\n"
                f"Adding `.recurrent_classes`\n"
                f"       `.transient_classes`\n"
                f"       `.is_irreducible`\n"
                f"    Finish",
                time=start,
            )
        else:
            logg.warning(
                "The transition matrix is irreducible, cannot further partition it\n    Finish",
                time=start,
            )

    @property
    def is_irreducible(self):
        """Whether the Markov chain is irreducible or not."""
        return self._is_irreducible

    @property
    def recurrent_classes(self):
        """Recurrent classes of the Markov chain."""
        return self._rec_classes

    @property
    def transient_classes(self):
        """Transient classes of the Markov chain."""
        return self._trans_classes


class LineageEstimatorMixin(TerminalStates, AbsProbs, LinDrivers, Pickleable, ABC):
    """Mixin containing terminal states and absorption probabilities."""



"""Lineages module."""

from typing import Union, TypeVar, Optional, Sequence

from cellrank import logging as logg
from cellrank.ul._docs import d
from cellrank.tl._utils import TestMethod
from cellrank.tl.kernels import PrecomputedKernel
from cellrank.tl._constants import AbsProbKey, TermStatesKey, TerminalStatesPlot
from cellrank.tl.estimators import GPCCA
from cellrank.tl.estimators._constants import P
from cellrank.tl.kernels._precomputed_kernel import DummyKernel

import pandas as pd

AnnData = TypeVar("AnnData")


@d.dedent
def lineages(
    adata: AnnData,
    backward: bool = False,
    copy: bool = False,
    return_estimator: bool = False,
    **kwargs,
) -> Optional[AnnData]:
    """
    Compute probabilistic lineage assignment using RNA velocity.

    For each cell `i` in :math:`{1, ..., N}` and %(initial_or_terminal)s state `j` in :math:`{1, ..., M}`,
    the probability is computed that cell `i` is either going to %(terminal)s state `j` (``backward=False``)
    or is coming from %(initial)s state `j` (``backward=True``).

    This function computes the absorption probabilities of a Markov chain towards the %(initial_or_terminal) states
    uncovered by :func:`cellrank.tl.initial_states` or :func:`cellrank.tl.terminal_states` using a highly efficient
    implementation that scales to large cell numbers.

    It's also possible to calculate mean and variance of the time until absorption for all or just a subset
    of the %(initial_or_terminal)s states. This can be seen as a pseudotemporal measure, either towards any terminal
    population of the state change trajectory, or towards specific ones.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    copy
        Whether to update the existing ``adata`` object or to return a copy.
    return_estimator
        Whether to return the estimator. Only available when ``copy=False``.
    kwargs
        Keyword arguments for :meth:`cellrank.tl.estimators.BaseEstimator.compute_absorption_probabilities`.

    Returns
    -------
    :class:`anndata.AnnData`, :class:`cellrank.tl.estimators.BaseEstimator` or :obj:`None`
        Depending on ``copy`` and ``return_estimator``, either updates the existing ``adata`` object,
        returns its copy or returns the estimator.
    """

    if backward:
        lin_key = AbsProbKey.BACKWARD
        fs_key = TermStatesKey.BACKWARD
        fs_key_pretty = TerminalStatesPlot.BACKWARD
    else:
        lin_key = AbsProbKey.FORWARD
        fs_key = TermStatesKey.FORWARD
        fs_key_pretty = TerminalStatesPlot.FORWARD

    try:
        pk = PrecomputedKernel(adata=adata, backward=backward)
    except KeyError as e:
        raise RuntimeError(
            f"Compute transition matrix first as `cellrank.tl.transition_matrix(..., backward={backward})`."
        ) from e

    start = logg.info(f"Computing lineage probabilities towards {fs_key_pretty.s}")
    mc = GPCCA(
        pk, read_from_adata=True, inplace=not copy
    )  # GPCCA is more general than CFLARE, in terms of what is saves
    if mc._get(P.TERM) is None:
        raise RuntimeError(
            f"Compute the states first as `cellrank.tl.{fs_key.s}(..., backward={backward})`."
        )

    # compute the absorption probabilities
    mc.compute_absorption_probabilities(**kwargs)

    logg.info(f"Adding lineages to `adata.obsm[{lin_key.s!r}]`\n    Finish", time=start)

    return mc.adata if copy else mc if return_estimator else None


@d.dedent
def lineage_drivers(
    adata: AnnData,
    backward: bool = False,
    lineages: Optional[Union[Sequence, str]] = None,
    method: str = TestMethod.FISCHER.s,
    cluster_key: Optional[str] = None,
    clusters: Optional[Union[Sequence, str]] = None,
    layer: str = "X",
    use_raw: bool = False,
    confidence_level: float = 0.95,
    n_perms: int = 1000,
    seed: Optional[int] = None,
    return_drivers: bool = True,
    **kwargs,
) -> Optional[pd.DataFrame]:
    """
    %(lineage_drivers.full_desc)s

    Parameters
    ----------
    %(adata)s
    %(backward)s
    %(lineage_drivers.parameters)s

    Returns
    -------
    %(lineage_drivers.returns)s

    References
    ----------
    %(lineage_drivers.references)s
    """  # noqa: D400

    # create dummy kernel and estimator
    pk = DummyKernel(adata, backward=backward)
    g = GPCCA(pk, read_from_adata=True, write_to_adata=False)
    if g._get(P.ABS_PROBS) is None:
        raise RuntimeError(
            f"Compute absorption probabilities first as `cellrank.tl.lineages(..., backward={backward})`."
        )

    # call the underlying function to compute and store the lineage drivers
    return g.compute_lineage_drivers(
        method=method,
        lineages=lineages,
        cluster_key=cluster_key,
        clusters=clusters,
        layer=layer,
        use_raw=use_raw,
        confidence_level=confidence_level,
        n_perms=n_perms,
        seed=seed,
        return_drivers=return_drivers,
        **kwargs,
    )



import os
from typing import Tuple
from tempfile import TemporaryDirectory

import pytest
from _helpers import assert_estimators_equal

import cellrank as cr
import cellrank.tl.kernels._precomputed_kernel
from anndata import AnnData
from cellrank.tl.kernels import VelocityKernel, ConnectivityKernel
from cellrank.tl._constants import (
    Direction,
    DirPrefix,
    AbsProbKey,
    TermStatesKey,
    _pd,
    _probs,
    _colors,
    _lin_names,
)
from cellrank.tl.estimators._constants import A, P

import numpy as np
import pandas as pd
from pandas.api.types import is_categorical_dtype

EPS = np.finfo(np.float64).eps


class TestCFLARE:
    def test_compute_partition(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_partition()

        assert isinstance(mc.is_irreducible, bool)
        if not mc.is_irreducible:
            assert isinstance(mc.recurrent_classes, list)
            assert isinstance(mc.transient_classes, list)
            assert f"{TermStatesKey.FORWARD}_rec_classes" in mc.adata.obs
            assert f"{TermStatesKey.FORWARD}_trans_classes" in mc.adata.obs
        else:
            assert mc.recurrent_classes is None
            assert mc.transient_classes is None

    def test_compute_eigendecomposition(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=2)

        assert isinstance(mc.eigendecomposition, dict)
        assert set(mc._get(P.EIG).keys()) == {
            "D",
            "V_l",
            "V_r",
            "eigengap",
            "params",
            "stationary_dist",
        }
        assert f"eig_{Direction.FORWARD}" in mc.adata.uns.keys()

    def test_compute_terminal_states_no_eig(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        with pytest.raises(RuntimeError):
            mc.compute_terminal_states(use=2)

    def test_compute_terminal_states_too_large_use(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=2)
        with pytest.raises(ValueError):
            mc.compute_terminal_states(use=1000)

    def test_compute_approx_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        assert is_categorical_dtype(mc._get(P.TERM))
        assert mc._get(P.TERM_PROBS) is not None

        assert TermStatesKey.FORWARD.s in mc.adata.obs.keys()
        assert _probs(TermStatesKey.FORWARD) in mc.adata.obs.keys()
        assert _colors(TermStatesKey.FORWARD) in mc.adata.uns.keys()

    def test_rename_terminal_states_no_terminal_states(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        with pytest.raises(RuntimeError):
            mc.rename_terminal_states({"foo": "bar"})

    def test_rename_terminal_states_invalid_old_name(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        with pytest.raises(ValueError):
            mc.rename_terminal_states({"foo": "bar"})

    def test_rename_terminal_states_invalid_new_name(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        with pytest.raises(ValueError):
            mc.rename_terminal_states({"0": "1"})

    def test_rename_terminal_states_try_joinining_states(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        with pytest.raises(ValueError):
            mc.rename_terminal_states({"0, 1": "foo"})

    def test_rename_terminal_states_empty_mapping(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        orig_cats = list(mc._get(P.TERM).cat.categories)
        mc.rename_terminal_states({})

        np.testing.assert_array_equal(mc._get(P.TERM).cat.categories, orig_cats)

    def test_rename_terminal_states_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        mc.rename_terminal_states({"0": "foo", "1": "bar"})

        np.testing.assert_array_equal(mc._get(P.TERM).cat.categories, ["foo", "bar"])

    def test_rename_terminal_states_non_string_new_names(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        mc.rename_terminal_states({"0": 42, "1": object})

        np.testing.assert_array_equal(
            mc._get(P.TERM).cat.categories, ["42", str(object)]
        )

    def test_rename_terminal_states_update_adata(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        mc.rename_terminal_states({"0": "foo", "1": "bar"})

        np.testing.assert_array_equal(mc._get(P.TERM).cat.categories, ["foo", "bar"])
        np.testing.assert_array_equal(
            mc.adata.obs[TermStatesKey.FORWARD.s].cat.categories, ["foo", "bar"]
        )
        np.testing.assert_array_equal(
            mc.adata.uns[_lin_names(TermStatesKey.FORWARD.s)], ["foo", "bar"]
        )

    def test_rename_terminal_states_dont_update_adata(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        mc.rename_terminal_states({"0": "foo", "1": "bar"}, update_adata=False)

        np.testing.assert_array_equal(mc._get(P.TERM).cat.categories, ["foo", "bar"])
        np.testing.assert_array_equal(
            mc.adata.obs[TermStatesKey.FORWARD.s].cat.categories, ["0", "1"]
        )
        np.testing.assert_array_equal(
            mc.adata.uns[_lin_names(TermStatesKey.FORWARD.s)], ["0", "1"]
        )

    def test_compute_absorption_probabilities_no_args(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        with pytest.raises(RuntimeError):
            mc.compute_absorption_probabilities()

    def test_compute_absorption_probabilities_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        assert isinstance(mc._get(P.PRIME_DEG), pd.Series)
        assert _pd(AbsProbKey.FORWARD) in mc.adata.obs.keys()
        np.testing.assert_array_equal(
            mc._get(P.PRIME_DEG), mc.adata.obs[_pd(AbsProbKey.FORWARD)]
        )

        assert isinstance(mc._get(P.ABS_PROBS), cr.tl.Lineage)
        assert mc._get(P.ABS_PROBS).shape == (mc.adata.n_obs, 2)
        assert f"{AbsProbKey.FORWARD}" in mc.adata.obsm.keys()
        np.testing.assert_array_equal(
            mc._get(P.ABS_PROBS).X, mc.adata.obsm[f"{AbsProbKey.FORWARD}"]
        )

        assert _lin_names(AbsProbKey.FORWARD) in mc.adata.uns.keys()
        np.testing.assert_array_equal(
            mc._get(P.ABS_PROBS).names,
            mc.adata.uns[_lin_names(AbsProbKey.FORWARD)],
        )

        assert _colors(AbsProbKey.FORWARD) in mc.adata.uns.keys()
        np.testing.assert_array_equal(
            mc._get(P.ABS_PROBS).colors,
            mc.adata.uns[_colors(AbsProbKey.FORWARD)],
        )
        np.testing.assert_allclose(mc._get(P.ABS_PROBS).X.sum(1), 1, rtol=1e-6)

    def test_compute_absorption_probabilities_solver(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        tol = 1e-6

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        # compute lin probs using direct solver
        mc.compute_absorption_probabilities(solver="direct")
        l_direct = mc._get(P.ABS_PROBS).copy()

        # compute lin probs using iterative solver
        mc.compute_absorption_probabilities(solver="gmres", tol=tol)
        l_iterative = mc._get(P.ABS_PROBS).copy()

        assert not np.shares_memory(l_direct.X, l_iterative.X)  # sanity check
        np.testing.assert_allclose(l_direct.X, l_iterative.X, rtol=0, atol=tol)

    def test_compute_absorption_probabilities_solver_petsc(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        tol = 1e-6

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        # compute lin probs using direct solver
        mc.compute_absorption_probabilities(solver="gmres", use_petsc=False, tol=tol)
        l_iter = mc._get(P.ABS_PROBS).copy()

        # compute lin probs using petsc iterative solver
        mc.compute_absorption_probabilities(solver="gmres", use_petsc=True, tol=tol)
        l_iter_petsc = mc._get(P.ABS_PROBS).copy()

        assert not np.shares_memory(l_iter.X, l_iter_petsc.X)  # sanity check
        np.testing.assert_allclose(l_iter.X, l_iter_petsc.X, rtol=0, atol=tol)

    def test_compute_absorption_probabilities_lineage_absorption_mean(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        tol = 1e-6

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        # compute lin probs using direct solver
        mc.compute_absorption_probabilities(
            solver="gmres",
            use_petsc=False,
            tol=tol,
            time_to_absorption="0",
        )
        at = mc._get(P.LIN_ABS_TIMES)

        assert isinstance(at, pd.DataFrame)
        np.testing.assert_array_equal(at.index, adata_large.obs_names)
        np.testing.assert_array_equal(at.columns, ["0 mean"])

    def test_compute_absorption_probabilities_lineage_absorption_var(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        tol = 1e-6

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        # compute lin probs using direct solver
        mc.compute_absorption_probabilities(
            solver="gmres", use_petsc=False, tol=tol, time_to_absorption={"0": "var"}
        )
        at = mc._get(P.LIN_ABS_TIMES)

        assert isinstance(at, pd.DataFrame)
        np.testing.assert_array_equal(at.index, adata_large.obs_names)
        np.testing.assert_array_equal(at.columns, ["0 mean", "0 var"])

    def test_compute_absorption_probabilities_lineage_absorption_all(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        tol = 1e-6

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)

        # compute lin probs using direct solver
        mc.compute_absorption_probabilities(
            solver="gmres", use_petsc=False, tol=tol, time_to_absorption={"all": "var"}
        )
        name = ", ".join(mc._get(P.ABS_PROBS).names)
        at = mc._get(P.LIN_ABS_TIMES)

        assert isinstance(at, pd.DataFrame)
        np.testing.assert_array_equal(at.index, adata_large.obs_names)
        np.testing.assert_array_equal(at.columns, [f"{name} mean", f"{name} var"])

    def test_compute_lineage_drivers_no_lineages(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        with pytest.raises(RuntimeError):
            mc.compute_lineage_drivers()

    def test_compute_lineage_drivers_invalid_lineages(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        with pytest.raises(KeyError):
            mc.compute_lineage_drivers(use_raw=False, lineages=["foo"])

    def test_compute_lineage_drivers_invalid_clusters(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        with pytest.raises(KeyError):
            mc.compute_lineage_drivers(
                use_raw=False, cluster_key="clusters", clusters=["foo"]
            )

    def test_compute_lineage_drivers_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        for lineage in ["0", "1"]:
            assert np.all(mc.adata.var[f"{DirPrefix.FORWARD} {lineage} corr"] >= -1.0)
            assert np.all(mc.adata.var[f"{DirPrefix.FORWARD} {lineage} corr"] <= 1.0)

            assert np.all(mc.adata.var[f"{DirPrefix.FORWARD} {lineage} qval"] >= 0)
            assert np.all(mc.adata.var[f"{DirPrefix.FORWARD} {lineage} qval"] <= 1.0)

    def test_plot_lineage_drivers_not_computed(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()

        with pytest.raises(RuntimeError):
            mc.plot_lineage_drivers("0")

    def test_plot_lineage_drivers_invalid_name(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        with pytest.raises(KeyError):
            mc.plot_lineage_drivers("foo", use_raw=False)

    def test_plot_lineage_drivers_invalid_n_genes(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        with pytest.raises(ValueError):
            mc.plot_lineage_drivers("0", use_raw=False, n_genes=0)

    def test_plot_lineage_drivers_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.compute_terminal_states(use=2)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")
        mc.plot_lineage_drivers("0", use_raw=False)

    def test_compute_absorption_probabilities_keys_colors(self, adata_large: AnnData):
        adata = adata_large
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc_fwd = cr.tl.estimators.CFLARE(terminal_kernel)
        mc_fwd.compute_partition()
        mc_fwd.compute_eigendecomposition()
        mc_fwd.compute_terminal_states(use=3)

        arcs = ["0", "2"]
        arc_colors = [
            c
            for arc, c in zip(
                mc_fwd._get(P.TERM).cat.categories, mc_fwd._get(A.TERM_COLORS)
            )
            if arc in arcs
        ]

        mc_fwd.compute_absorption_probabilities(keys=arcs)
        lin_colors = mc_fwd._get(P.ABS_PROBS)[arcs].colors

        np.testing.assert_array_equal(arc_colors, lin_colors)

    def compare_absorption_probabilites_with_reference(self):
        # define a reference transition matrix. This is an absorbing MC with 2 absorbing states
        transition_matrix = np.array(
            [
                # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.
                [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0.
                [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1.
                [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2.
                [0.0, 0.05, 0.05, 0.0, 0.45, 0.45, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3.
                [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0],  # 4.
                [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.4, 0.0, 0.0],  # 5.
                [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0],  # 6.
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],  # 7.
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0],  # 8.
                [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2],  # 9.
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],  # 10.
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0],  # 11.
            ]
        )

        absorption_probabilities_reference = np.array(
            [
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.5, 0.5],
                [0.60571429, 0.39428571],
                [0.39428571, 0.60571429],
                [0.94047619, 0.05952381],
                [0.95238095, 0.04761905],
                [0.05952381, 0.94047619],
                [0.04761905, 0.95238095],
            ]
        )

        # initialise a pre-computed kernel and CFLARE estimator object
        c = cr.tl.estimators.CFLARE(
            cellrank.tl.kernels._precomputed_kernel.PrecomputedKernel(transition_matrix)
        )

        # define the set of macrostates
        state_annotation = pd.Series(index=range(p.shape[0]))
        state_annotation[7] = "terminal_1"
        state_annotation[10] = "terminal_2"
        state_annotation = state_annotation.astype("category")
        c._set(A.TERM, state_annotation)

        # compute absorption probabilities
        c.compute_absorption_probabilities()
        absorption_probabilities_query = c.get(P.ABS_PROBS)[state_annotation.isna()]

        # check whether these two agree
        np.allclose(absorption_probabilities_query, absorption_probabilities_reference)

    def test_manual_approx_rc_set(self, adata_large):
        adata = adata_large
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc_fwd = cr.tl.estimators.CFLARE(terminal_kernel)
        mc_fwd.compute_partition()
        mc_fwd.compute_eigendecomposition()

        mc_fwd.compute_terminal_states(use=3)
        original = np.array(adata.obs[f"{TermStatesKey.FORWARD}"].copy())
        zero_mask = original == "0"

        cells = list(adata[zero_mask].obs_names)
        mc_fwd.set_terminal_states({"foo": cells})

        assert (adata.obs[f"{TermStatesKey.FORWARD}"][zero_mask] == "foo").all()
        assert pd.isna(adata.obs[f"{TermStatesKey.FORWARD}"][~zero_mask]).all()

    def test_abs_probs_do_not_sum_to_1(self, adata_large: AnnData, mocker):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.set_terminal_states(
            {"x": adata_large.obs_names[:3], "y": adata_large.obs_names[3:6]}
        )

        n_term = np.sum(~pd.isnull(mc.terminal_states))
        abs_prob = np.zeros((adata_large.n_obs - n_term, n_term))
        abs_prob[:, 0] = 1.0
        abs_prob[0, 0] = 1.01
        mocker.patch(
            "cellrank.tl.estimators._base_estimator._solve_lin_system",
            return_value=abs_prob,
        )

        with pytest.raises(
            ValueError, match=r"`1` value\(s\) do not sum to 1 \(rtol=1e-3\)."
        ):
            mc.compute_absorption_probabilities()

    def test_abs_probs_negative(self, adata_large: AnnData, mocker):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.CFLARE(terminal_kernel)
        mc.compute_eigendecomposition(k=5)
        mc.set_terminal_states(
            {"x": adata_large.obs_names[:3], "y": adata_large.obs_names[3:6]}
        )

        n_term = np.sum(~pd.isnull(mc.terminal_states))
        abs_prob = np.zeros((adata_large.n_obs - n_term, n_term))
        abs_prob[:, 0] = 1.0
        abs_prob[0, 0] = -0.5
        abs_prob[0, 1] = -1.5
        mocker.patch(
            "cellrank.tl.estimators._base_estimator._solve_lin_system",
            return_value=abs_prob,
        )

        with pytest.raises(ValueError, match=r"`2` value\(s\) are negative."):
            mc.compute_absorption_probabilities()


class TestCFLAREIO:
    def test_copy(self, adata_cflare_fwd: Tuple[AnnData, cr.tl.estimators.CFLARE]):
        _, mc1 = adata_cflare_fwd
        mc2 = mc1.copy()

        assert_estimators_equal(mc1, mc2, copy=True)

    def test_read(self, adata_cflare_fwd: Tuple[AnnData, cr.tl.estimators.CFLARE]):
        _, mc1 = adata_cflare_fwd

        with TemporaryDirectory() as tmpdir:
            mc1.write(os.path.join(tmpdir, "foo"))
            mc2 = cr.tl.estimators.CFLARE.read(os.path.join(tmpdir, "foo.pickle"))

        assert_estimators_equal(mc1, mc2)
